# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import base64
import copy
import json
from textwrap import dedent

import pytest

from cloudinit import util
from cloudinit.sources import DataSourceIBMCloud as ibm
from tests.unittests import helpers as test_helpers

mock = test_helpers.mock

D_PATH = "cloudinit.sources.DataSourceIBMCloud."


@mock.patch(D_PATH + "_is_xen", return_value=True)
@mock.patch(D_PATH + "_is_ibm_provisioning")
@mock.patch(D_PATH + "util.blkid")
class TestGetIBMPlatform:
    """Test the get_ibm_platform helper."""

    blkid_base = {
        "/dev/xvda1": {
            "DEVNAME": "/dev/xvda1",
            "LABEL": "cloudimg-bootfs",
            "TYPE": "ext3",
        },
        "/dev/xvda2": {
            "DEVNAME": "/dev/xvda2",
            "LABEL": "cloudimg-rootfs",
            "TYPE": "ext4",
        },
    }

    blkid_metadata_disk = {
        "/dev/xvdh1": {
            "DEVNAME": "/dev/xvdh1",
            "LABEL": "METADATA",
            "TYPE": "vfat",
            "SEC_TYPE": "msdos",
            "UUID": "681B-8C5D",
            "PARTUUID": "3d631e09-01",
        },
    }

    blkid_oscode_disk = {
        "/dev/xvdh": {
            "DEVNAME": "/dev/xvdh",
            "LABEL": "config-2",
            "TYPE": "vfat",
            "SEC_TYPE": "msdos",
            "UUID": ibm.IBM_CONFIG_UUID,
        }
    }

    @pytest.fixture(autouse=True)
    def fixtures(self):
        self.blkid_metadata = copy.deepcopy(self.blkid_base)
        self.blkid_metadata.update(copy.deepcopy(self.blkid_metadata_disk))

        self.blkid_oscode = copy.deepcopy(self.blkid_base)
        self.blkid_oscode.update(copy.deepcopy(self.blkid_oscode_disk))

    def test_id_template_live_metadata(self, m_blkid, m_is_prov, _m_xen):
        """identify TEMPLATE_LIVE_METADATA."""
        m_blkid.return_value = self.blkid_metadata
        m_is_prov.return_value = False
        assert (
            ibm.Platforms.TEMPLATE_LIVE_METADATA,
            "/dev/xvdh1",
        ) == ibm.get_ibm_platform()

    def test_id_template_prov_metadata(self, m_blkid, m_is_prov, _m_xen):
        """identify TEMPLATE_PROVISIONING_METADATA."""
        m_blkid.return_value = self.blkid_metadata
        m_is_prov.return_value = True
        assert (
            ibm.Platforms.TEMPLATE_PROVISIONING_METADATA,
            "/dev/xvdh1",
        ) == ibm.get_ibm_platform()

    def test_id_template_prov_nodata(self, m_blkid, m_is_prov, _m_xen):
        """identify TEMPLATE_PROVISIONING_NODATA."""
        m_blkid.return_value = self.blkid_base
        m_is_prov.return_value = True
        assert (
            ibm.Platforms.TEMPLATE_PROVISIONING_NODATA,
            None,
        ) == ibm.get_ibm_platform()

    def test_id_os_code(self, m_blkid, m_is_prov, _m_xen):
        """Identify OS_CODE."""
        m_blkid.return_value = self.blkid_oscode
        m_is_prov.return_value = False
        assert (ibm.Platforms.OS_CODE, "/dev/xvdh") == ibm.get_ibm_platform()

    def test_id_os_code_must_match_uuid(self, m_blkid, m_is_prov, _m_xen):
        """Test against false positive on openstack with non-ibm UUID."""
        blkid = self.blkid_oscode
        blkid["/dev/xvdh"]["UUID"] = "9999-9999"
        m_blkid.return_value = blkid
        m_is_prov.return_value = False
        assert (None, None) == ibm.get_ibm_platform()


@mock.patch(D_PATH + "_read_system_uuid", return_value=None)
@mock.patch(D_PATH + "get_ibm_platform")
class TestReadMD:
    """Test the read_datasource helper."""

    template_md = {
        "files": [],
        "network_config": {"content_path": "/content/interfaces"},
        "hostname": "ci-fond-ram",
        "name": "ci-fond-ram",
        "domain": "testing.ci.cloud-init.org",
        "meta": {"dsmode": "net"},
        "uuid": "8e636730-9f5d-c4a5-327c-d7123c46e82f",
        "public_keys": {"1091307": "ssh-rsa AAAAB3NzaC1...Hw== ci-pubkey"},
    }

    oscode_md = {
        "hostname": "ci-grand-gannet.testing.ci.cloud-init.org",
        "name": "ci-grand-gannet",
        "uuid": "2f266908-8e6c-4818-9b5c-42e9cc66a785",
        "random_seed": "bm90LXJhbmRvbQo=",
        "crypt_key": "ssh-rsa AAAAB3NzaC1yc2..n6z/",
        "configuration_token": "eyJhbGciOi..M3ZA",
        "public_keys": {"1091307": "ssh-rsa AAAAB3N..Hw== ci-pubkey"},
    }

    content_interfaces = dedent(
        """\
        auto lo
        iface lo inet loopback

        auto eth0
        allow-hotplug eth0
        iface eth0 inet static
        address 10.82.43.5
        netmask 255.255.255.192
        """
    )

    userdata = b"#!/bin/sh\necho hi mom\n"
    # meta.js file gets json encoded userdata as a list.
    meta_js = '["#!/bin/sh\necho hi mom\n"]'
    vendor_data = {
        "cloud-init": "#!/bin/bash\necho 'root:$6$5ab01p1m1' | chpasswd -e"
    }

    network_data = {
        "links": [
            {
                "id": "interface_29402281",
                "name": "eth0",
                "mtu": None,
                "type": "phy",
                "ethernet_mac_address": "06:00:f1:bd:da:25",
            },
            {
                "id": "interface_29402279",
                "name": "eth1",
                "mtu": None,
                "type": "phy",
                "ethernet_mac_address": "06:98:5e:d0:7f:86",
            },
        ],
        "networks": [
            {
                "id": "network_109887563",
                "link": "interface_29402281",
                "type": "ipv4",
                "ip_address": "10.82.43.2",
                "netmask": "255.255.255.192",
                "routes": [
                    {
                        "network": "10.0.0.0",
                        "netmask": "255.0.0.0",
                        "gateway": "10.82.43.1",
                    },
                    {
                        "network": "161.26.0.0",
                        "netmask": "255.255.0.0",
                        "gateway": "10.82.43.1",
                    },
                ],
            },
            {
                "id": "network_109887551",
                "link": "interface_29402279",
                "type": "ipv4",
                "ip_address": "108.168.194.252",
                "netmask": "255.255.255.248",
                "routes": [
                    {
                        "network": "0.0.0.0",
                        "netmask": "0.0.0.0",
                        "gateway": "108.168.194.249",
                    }
                ],
            },
        ],
        "services": [
            {"type": "dns", "address": "10.0.80.11"},
            {"type": "dns", "address": "10.0.80.12"},
        ],
    }

    sysuuid = "7f79ebf5-d791-43c3-a723-854e8389d59f"

    def _get_expected_metadata(self, os_md):
        """return expected 'metadata' for data loaded from meta_data.json."""
        os_md = copy.deepcopy(os_md)
        renames = (
            ("hostname", "local-hostname"),
            ("uuid", "instance-id"),
            ("public_keys", "public-keys"),
        )
        ret = {}
        for osname, mdname in renames:
            if osname in os_md:
                ret[mdname] = os_md[osname]
        if "random_seed" in os_md:
            ret["random_seed"] = base64.b64decode(os_md["random_seed"])

        return ret

    def test_provisioning_md(self, m_platform, m_sysuuid):
        """Provisioning env with a metadata disk should return None."""
        m_platform.return_value = (
            ibm.Platforms.TEMPLATE_PROVISIONING_METADATA,
            "/dev/xvdh",
        )
        assert ibm.read_md() is None

    def test_provisioning_no_metadata(self, m_platform, m_sysuuid):
        """Provisioning env with no metadata disk should return None."""
        m_platform.return_value = (
            ibm.Platforms.TEMPLATE_PROVISIONING_NODATA,
            None,
        )
        assert ibm.read_md() is None

    def test_provisioning_not_ibm(self, m_platform, m_sysuuid):
        """Provisioning env but not identified as IBM should return None."""
        m_platform.return_value = (None, None)
        assert ibm.read_md() is None

    def test_template_live(self, m_platform, m_sysuuid, tmp_path):
        """Template live environment should be identified."""
        tmpdir = str(tmp_path)
        m_platform.return_value = (
            ibm.Platforms.TEMPLATE_LIVE_METADATA,
            tmpdir,
        )
        m_sysuuid.return_value = self.sysuuid

        test_helpers.populate_dir(
            tmpdir,
            {
                "openstack/latest/meta_data.json": json.dumps(
                    self.template_md
                ),
                "openstack/latest/user_data": self.userdata,
                "openstack/content/interfaces": self.content_interfaces,
                "meta.js": self.meta_js,
            },
        )

        ret = ibm.read_md()
        if ret is None:  # this is needed for mypy - ensures ret is not None
            pytest.fail("read_md returned None unexpectedly")
        assert ibm.Platforms.TEMPLATE_LIVE_METADATA == ret["platform"]
        assert tmpdir == ret["source"]
        assert self.userdata == ret["userdata"]
        assert self._get_expected_metadata(self.template_md) == ret["metadata"]
        assert self.sysuuid == ret["system-uuid"]

    def test_os_code_live(self, m_platform, m_sysuuid, tmp_path):
        """Verify an os_code metadata path."""
        tmpdir = str(tmp_path)
        m_platform.return_value = (ibm.Platforms.OS_CODE, tmpdir)
        netdata = json.dumps(self.network_data)
        test_helpers.populate_dir(
            tmpdir,
            {
                "openstack/latest/meta_data.json": json.dumps(self.oscode_md),
                "openstack/latest/user_data": self.userdata,
                "openstack/latest/vendor_data.json": json.dumps(
                    self.vendor_data
                ),
                "openstack/latest/network_data.json": netdata,
            },
        )

        ret = ibm.read_md()
        if ret is None:  # this is needed for mypy - ensures ret is not None
            pytest.fail("read_md returned None unexpectedly")
        assert ibm.Platforms.OS_CODE == ret["platform"]
        assert tmpdir == ret["source"]
        assert self.userdata == ret["userdata"]
        assert self._get_expected_metadata(self.oscode_md) == ret["metadata"]

    def test_os_code_live_no_userdata(self, m_platform, m_sysuuid, tmp_path):
        """Verify os_code without user-data."""
        tmpdir = str(tmp_path)
        m_platform.return_value = (ibm.Platforms.OS_CODE, tmpdir)
        test_helpers.populate_dir(
            tmpdir,
            {
                "openstack/latest/meta_data.json": json.dumps(self.oscode_md),
                "openstack/latest/vendor_data.json": json.dumps(
                    self.vendor_data
                ),
            },
        )

        ret = ibm.read_md()
        if ret is None:  # this is needed for mypy - ensures ret is not None
            pytest.fail("read_md returned None unexpectedly")
        assert ibm.Platforms.OS_CODE == ret["platform"]
        assert tmpdir == ret["source"]
        assert ret["userdata"] is None
        assert self._get_expected_metadata(self.oscode_md) == ret["metadata"]


@pytest.mark.usefixtures("fake_filesystem")
class TestIsIBMProvisioning:
    """Test the _is_ibm_provisioning method."""

    inst_log = "/root/swinstall.log"
    prov_cfg = "/root/provisioningConfiguration.cfg"
    boot_ref = "/proc/1/environ"

    def test_no_config(self):
        """No provisioning config means not provisioning."""
        assert not ibm._is_ibm_provisioning()

    def test_config_only(self, tmp_path):
        """A provisioning config without a log means provisioning."""
        test_helpers.populate_dir(str(tmp_path), {self.prov_cfg: "key=value"})
        assert ibm._is_ibm_provisioning()

    def test_config_with_old_log(self, caplog, tmp_path):
        """A config with a log from previous boot is not provisioning."""
        data = {
            self.prov_cfg: ("key=value\nkey2=val2\n", -10),
            self.inst_log: ("log data\n", -30),
            self.boot_ref: ("PWD=/", 0),
        }
        test_helpers.populate_dir_with_ts(str(tmp_path), data)
        assert not ibm._is_ibm_provisioning()
        assert "from previous boot" in caplog.text

    def test_config_with_new_log(self, caplog, tmp_path):
        """A config with a log from this boot is provisioning."""
        data = {
            self.prov_cfg: ("key=value\nkey2=val2\n", -10),
            self.inst_log: ("log data\n", 30),
            self.boot_ref: ("PWD=/", 0),
        }
        test_helpers.populate_dir_with_ts(str(tmp_path), data)
        assert ibm._is_ibm_provisioning()
        assert "from current boot" in caplog.text

    def test_config_and_log_no_reference(self, caplog, tmp_path):
        """If the config and log existed, but no reference, assume not."""
        test_helpers.populate_dir(
            str(tmp_path),
            {self.prov_cfg: "key=value", self.inst_log: "log data\n"},
        )
        assert not ibm._is_ibm_provisioning()
        assert "no reference file" in caplog.text


class TestDataSourceIBMCloud:
    @pytest.fixture(autouse=True)
    def fixture(self, paths):
        util.ensure_dir(paths.cloud_dir)
        self.ds = ibm.DataSourceIBMCloud(sys_cfg={}, distro=None, paths=paths)

    def test_get_data_false(self):
        """When read_md returns None, get_data returns False."""
        with mock.patch(D_PATH + "read_md", return_value=None):
            assert not self.ds.get_data()

    def test_get_data_processes_read_md(self):
        """get_data processes and caches content returned by read_md."""
        md = {
            "metadata": {},
            "networkdata": "net",
            "platform": "plat",
            "source": "src",
            "system-uuid": "uuid",
            "userdata": "ud",
            "vendordata": "vd",
        }
        with mock.patch(D_PATH + "read_md", return_value=md):
            assert self.ds.get_data()
        assert "src" == self.ds.source
        assert "plat" == self.ds.platform
        assert {} == self.ds.metadata
        assert "ud" == self.ds.userdata_raw
        assert "net" == self.ds.network_json
        assert "uuid" == self.ds.system_uuid
        assert "ibmcloud" == self.ds.cloud_name
        assert "ibmcloud" == self.ds.platform_type
        assert "plat (src)" == self.ds.subplatform



# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""SSH AuthKey Fingerprints: Log fingerprints of user SSH keys"""

import base64
import hashlib
import logging

from cloudinit import ssh_util, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.distros import ALL_DISTROS, ug_util
from cloudinit.log import log_util
from cloudinit.settings import PER_INSTANCE
from cloudinit.simpletable import SimpleTable

meta: MetaSchema = {
    "id": "cc_ssh_authkey_fingerprints",
    "distros": [ALL_DISTROS],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": [],
}

LOG = logging.getLogger(__name__)


def _split_hash(bin_hash):
    split_up = []
    for i in range(0, len(bin_hash), 2):
        split_up.append(bin_hash[i : i + 2])
    return split_up


def _gen_fingerprint(b64_text, hash_meth="sha256"):
    if not b64_text:
        return ""
    # TBD(harlowja): Maybe we should feed this into 'ssh -lf'?
    try:
        hasher = hashlib.new(hash_meth)
        hasher.update(base64.b64decode(b64_text))
        return ":".join(_split_hash(hasher.hexdigest()))
    except (TypeError, ValueError):
        # Raised when b64 not really b64...
        # or when the hash type is not really
        # a known/supported hash type...
        return "?"


def _is_printable_key(entry):
    if any([entry.keytype, entry.base64, entry.comment, entry.options]):
        if (
            entry.keytype
            and entry.keytype.lower().strip() in ssh_util.VALID_KEY_TYPES
        ):
            return True
    return False


def _pprint_key_entries(
    user, key_fn, key_entries, hash_meth="sha256", prefix="ci-info: "
):
    if not key_entries:
        message = (
            "%sno authorized SSH keys fingerprints found for user %s.\n"
            % (prefix, user)
        )
        log_util.multi_log(message, console=True, stderr=False)
        return
    tbl_fields = [
        "Keytype",
        "Fingerprint (%s)" % (hash_meth),
        "Options",
        "Comment",
    ]
    tbl = SimpleTable(tbl_fields)
    for entry in key_entries:
        if _is_printable_key(entry):
            row = [
                entry.keytype or "-",
                _gen_fingerprint(entry.base64, hash_meth) or "-",
                entry.options or "-",
                entry.comment or "-",
            ]
            tbl.add_row(row)
    authtbl_s = tbl.get_string()
    authtbl_lines = authtbl_s.splitlines()
    max_len = len(max(authtbl_lines, key=len))
    lines = [
        util.center(
            "Authorized keys from %s for user %s" % (key_fn, user),
            "+",
            max_len,
        ),
    ]
    lines.extend(authtbl_lines)
    for line in lines:
        log_util.multi_log(
            text="%s%s\n" % (prefix, line), stderr=False, console=True
        )


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    if util.is_true(cfg.get("no_ssh_fingerprints", False)):
        LOG.debug(
            "Skipping module named %s, logging of SSH fingerprints disabled",
            name,
        )
        return

    hash_meth = util.get_cfg_option_str(cfg, "authkey_hash", "sha256")
    (users, _groups) = ug_util.normalize_users_groups(cfg, cloud.distro)
    for user_name, _cfg in users.items():
        if _cfg.get("no_create_home") or _cfg.get("system"):
            LOG.debug(
                "Skipping printing of ssh fingerprints for user '%s' because "
                "no home directory is created",
                user_name,
            )
            continue

        (key_fn, key_entries) = ssh_util.extract_authorized_keys(user_name)
        _pprint_key_entries(user_name, key_fn, key_entries, hash_meth)



"""Integration tests for cc_ca_certs.

(This is ported from ``tests/cloud_tests//testcases/modules/ca_certs.yaml``.)

TODO:
* Mark this as running on Debian and Alpine (once we have marks for that)
* Implement testing for the RHEL-specific paths
"""

import os.path
from textwrap import indent

import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.releases import IS_UBUNTU
from tests.integration_tests.util import verify_clean_boot, verify_clean_log

CERT_CONTENT = """\
-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----"""

USER_DATA = f"""\
#cloud-config
ca_certs:
  remove_defaults: true
  trusted:
    - |
{indent(CERT_CONTENT, "      ")}
"""


@pytest.mark.skipif(
    not IS_UBUNTU, reason="CA cert functionality is distro specific"
)
@pytest.mark.user_data(USER_DATA)
class TestCaCerts:
    def test_certs_updated(self, class_client: IntegrationInstance):
        """Test that /etc/ssl/certs is updated as we expect."""
        root = "/etc/ssl/certs"
        filenames = class_client.execute(["ls", "-1", root]).splitlines()
        unlinked_files = []
        links = {}
        for filename in filenames:
            full_path = os.path.join(root, filename)
            symlink_target = class_client.execute(["readlink", full_path])
            is_symlink = symlink_target.ok
            if is_symlink:
                links[filename] = symlink_target
            else:
                unlinked_files.append(filename)

        assert ["ca-certificates.crt"] == unlinked_files
        assert "cloud-init-ca-cert-1.pem" == links["a535c1f3.0"]
        assert (
            "/usr/local/share/ca-certificates/cloud-init-ca-cert-1.crt"
            == links["cloud-init-ca-cert-1.pem"]
        )

    def test_cert_installed(self, class_client: IntegrationInstance):
        """Test that our specified cert has been installed"""
        checksum = class_client.execute(
            "sha256sum /etc/ssl/certs/ca-certificates.crt"
        )
        assert (
            "78e875f18c73c1aab9167ae0bd323391e52222cc2dbcda42d129537219300062"
            in checksum
        )

    def test_clean_log(self, class_client: IntegrationInstance):
        """Verify no errors or deprecations in log."""
        log = class_client.read_from_file("/var/log/cloud-init-output.log")
        verify_clean_log(log, ignore_deprecations=False)
        verify_clean_boot(class_client)



# This file is part of cloud-init. See LICENSE file for license information.

import pytest

from cloudinit import util
from tests.unittests.helpers import populate_dir


class TestPathPrefix2Dict:
    def test_required_only(self, tmp_path):
        dirdata = {"f1": b"f1content", "f2": b"f2content"}
        populate_dir(str(tmp_path), dirdata)

        ret = util.pathprefix2dict(str(tmp_path), required=["f1", "f2"])
        assert dirdata == ret

    def test_required_missing(self, tmp_path):
        dirdata = {"f1": b"f1content"}
        populate_dir(str(tmp_path), dirdata)
        kwargs = {"required": ["f1", "f2"]}
        with pytest.raises(ValueError):
            util.pathprefix2dict(str(tmp_path), **kwargs)

    def test_no_required_and_optional(self, tmp_path):
        dirdata = {"f1": b"f1c", "f2": b"f2c"}
        populate_dir(str(tmp_path), dirdata)

        ret = util.pathprefix2dict(
            str(tmp_path), required=None, optional=["f1", "f2"]
        )
        assert dirdata == ret

    def test_required_and_optional(self, tmp_path):
        dirdata = {"f1": b"f1c", "f2": b"f2c"}
        populate_dir(str(tmp_path), dirdata)

        ret = util.pathprefix2dict(
            str(tmp_path), required=["f1"], optional=["f2"]
        )
        assert dirdata == ret



# This file is part of cloud-init. See LICENSE file for license information.

"""Module for ephemeral network context managers"""
import contextlib
import logging
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple

import cloudinit.net as net
import cloudinit.netinfo as netinfo
from cloudinit.net.dhcp import NoDHCPLeaseError, maybe_perform_dhcp_discovery
from cloudinit.subp import ProcessExecutionError
from cloudinit.url_helper import UrlError, wait_for_url

LOG = logging.getLogger(__name__)


class EphemeralIPv4Network:
    """Context manager which sets up temporary static network configuration.

    No operations are performed if the provided interface already has the
    specified configuration.
    This can be verified with the connectivity_urls_data.
    If unconnected, bring up the interface with valid ip, prefix and broadcast.
    If router is provided setup a default route for that interface. Upon
    context exit, clean up the interface leaving no configuration behind.
    """

    def __init__(
        self,
        distro,
        interface,
        ip,
        prefix_or_mask,
        broadcast,
        interface_addrs_before_dhcp: dict,
        router=None,
        static_routes=None,
    ):
        """Setup context manager and validate call signature.

        @param interface: Name of the network interface to bring up.
        @param ip: IP address to assign to the interface.
        @param prefix_or_mask: Either netmask of the format X.X.X.X or an int
            prefix.
        @param broadcast: Broadcast address for the IPv4 network.
        @param router: Optionally the default gateway IP.
        @param static_routes: Optionally a list of static routes from DHCP
        """
        if not all([interface, ip, prefix_or_mask, broadcast]):
            raise ValueError(
                "Cannot init network on {0} with {1}/{2} and bcast {3}".format(
                    interface, ip, prefix_or_mask, broadcast
                )
            )
        try:
            self.prefix = net.ipv4_mask_to_net_prefix(prefix_or_mask)
        except ValueError as e:
            raise ValueError(
                "Cannot setup network, invalid prefix or "
                "netmask: {0}".format(e)
            ) from e

        self.interface = interface
        self.ip = ip
        self.broadcast = broadcast
        self.router = router
        self.static_routes = static_routes
        # List of commands to run to cleanup state.
        self.cleanup_cmds: List[Callable] = []
        self.distro = distro
        self.cidr = f"{self.ip}/{self.prefix}"
        self.interface_addrs_before_dhcp = interface_addrs_before_dhcp.get(
            self.interface, {}
        )

    def __enter__(self):
        """Set up ephemeral network if interface is not connected.

        This context manager handles the lifecycle of the network interface,
        addresses, routes, etc
        """

        try:
            try:
                self._bringup_device()
            except ProcessExecutionError as e:
                if "File exists" not in str(
                    e.stderr
                ) and "Address already assigned" not in str(e.stderr):
                    raise

            # rfc3442 requires us to ignore the router config *if*
            # classless static routes are provided.
            #
            # https://tools.ietf.org/html/rfc3442
            #
            # If the DHCP server returns both a Classless Static Routes
            # option and a Router option, the DHCP client MUST ignore
            # the Router option.
            #
            # Similarly, if the DHCP server returns both a Classless
            # Static Routes option and a Static Routes option, the DHCP
            # client MUST ignore the Static Routes option.
            if self.static_routes:
                self._bringup_static_routes()
            elif self.router:
                self._bringup_router()
        except ProcessExecutionError:
            self.__exit__(None, None, None)
            raise

    def __exit__(self, excp_type, excp_value, excp_traceback):
        """Teardown anything we set up."""
        for cmd in self.cleanup_cmds:
            cmd()

    def _bringup_device(self):
        """Perform the ip commands to fully set up the device.

        Dhcp clients behave differently in how they leave link state and ip
        address assignment.

        Attempt assigning address and setting up link if needed to be done.
        Set cleanup_cmds to return the interface state to how it was prior
        to execution of the dhcp client.
        """
        LOG.debug(
            "Attempting setup of ephemeral network on %s with %s brd %s",
            self.interface,
            self.cidr,
            self.broadcast,
        )
        interface_addrs_after_dhcp = netinfo.netdev_info().get(
            self.interface, {}
        )
        has_link = interface_addrs_after_dhcp.get("up")
        had_link = self.interface_addrs_before_dhcp.get("up")
        has_ip = self.ip in [
            ip.get("ip") for ip in interface_addrs_after_dhcp.get("ipv4", {})
        ]
        had_ip = self.ip in [
            ip.get("ip")
            for ip in self.interface_addrs_before_dhcp.get("ipv4", {})
        ]

        if has_ip:
            LOG.debug(
                "Skip adding ip address: %s already has address %s",
                self.interface,
                self.ip,
            )
        else:
            self.distro.net_ops.add_addr(
                self.interface, self.cidr, self.broadcast
            )
        if has_link:
            LOG.debug(
                "Skip bringing up network link: interface %s is already up",
                self.interface,
            )
        else:
            self.distro.net_ops.link_up(self.interface, family="inet")
        if had_link:
            LOG.debug(
                "Not queueing link down: link [%s] was up prior before "
                "receiving a dhcp lease",
                self.interface,
            )
        else:
            self.cleanup_cmds.append(
                partial(
                    self.distro.net_ops.link_down,
                    self.interface,
                    family="inet",
                )
            )
        if had_ip:
            LOG.debug(
                "Not queueing address removal: address %s was assigned before "
                "receiving a dhcp lease",
                self.ip,
            )
        else:
            self.cleanup_cmds.append(
                partial(
                    self.distro.net_ops.del_addr, self.interface, self.cidr
                )
            )

    def _bringup_static_routes(self):
        # static_routes = [("169.254.169.254/32", "130.56.248.255"),
        #                  ("0.0.0.0/0", "130.56.240.1")]
        for net_address, gateway in self.static_routes:
            # Use "append" rather than "add" since the DHCP server may provide
            # rfc3442 classless static routes with multiple routes to the same
            # subnet via different routers or local interface addresses.
            #
            # In this scenario, `ip r add` fails.
            #
            # RHBZ: #2003231
            self.distro.net_ops.append_route(
                self.interface, net_address, gateway
            )
            self.cleanup_cmds.insert(
                0,
                partial(
                    self.distro.net_ops.del_route,
                    self.interface,
                    net_address,
                    gateway=gateway,
                ),
            )

    def _bringup_router(self):
        """Perform the ip commands to fully setup the router if needed."""
        # Check if a default route exists and exit if it does
        out = self.distro.net_ops.get_default_route()
        if "default" in out:
            LOG.debug(
                "Skip ephemeral route setup. %s already has default route: %s",
                self.interface,
                out.strip(),
            )
            return
        self.distro.net_ops.add_route(
            self.interface, self.router, source_address=self.ip
        )
        self.cleanup_cmds.insert(
            0,
            partial(
                self.distro.net_ops.del_route,
                self.interface,
                self.router,
                source_address=self.ip,
            ),
        )
        self.distro.net_ops.add_route(
            self.interface, "default", gateway=self.router
        )
        self.cleanup_cmds.insert(
            0,
            partial(self.distro.net_ops.del_route, self.interface, "default"),
        )


class EphemeralIPv6Network:
    """Context manager which sets up a ipv6 link local address

    The linux kernel assigns link local addresses on link-up, which is
    sufficient for link-local communication.
    """

    def __init__(self, distro, interface):
        """Setup context manager and validate call signature.

        @param interface: Name of the network interface to bring up.
        @param ip: IP address to assign to the interface.
        @param prefix: IPv6 uses prefixes, not netmasks
        """
        if not interface:
            raise ValueError("Cannot init network on {0}".format(interface))

        self.interface = interface
        self.distro = distro

    def __enter__(self):
        """linux kernel does autoconfiguration even when autoconf=0

        https://www.kernel.org/doc/html/latest/networking/ipv6.html
        """
        if net.read_sys_net(self.interface, "operstate") != "up":
            self.distro.net_ops.link_up(self.interface)

    def __exit__(self, *_args):
        """No need to set the link to down state"""


class EphemeralDHCPv4:
    def __init__(
        self,
        distro,
        iface=None,
        connectivity_urls_data: Optional[List[Dict[str, Any]]] = None,
        dhcp_log_func: Optional[Callable[[str, str, str], None]] = None,
    ):
        self.iface = iface
        self._ephipv4: Optional[EphemeralIPv4Network] = None
        self.lease: Optional[Dict[str, Any]] = None
        self.dhcp_log_func = dhcp_log_func
        self.connectivity_urls_data = connectivity_urls_data or []
        self.distro = distro
        self.interface_addrs_before_dhcp = netinfo.netdev_info()

    def __enter__(self):
        """Setup sandboxed dhcp context, unless connectivity_url can already be
        reached."""
        if imds_reached_at_url := _check_connectivity_to_imds(
            self.connectivity_urls_data
        ):
            LOG.debug(
                "Skip ephemeral DHCP setup, instance has connectivity"
                " to %s",
                imds_reached_at_url,
            )
            return None
        # If we don't have connectivity, perform dhcp discovery
        return self.obtain_lease()

    def __exit__(self, excp_type, excp_value, excp_traceback):
        """Teardown sandboxed dhcp context."""
        self.clean_network()

    def clean_network(self):
        """Exit _ephipv4 context to teardown of ip configuration performed."""
        self.lease = None
        if self._ephipv4:
            self._ephipv4.__exit__(None, None, None)

    def obtain_lease(self):
        """Perform dhcp discovery in a sandboxed environment if possible.

        @return: A dict representing dhcp options on the most recent lease
            obtained from the dhclient discovery if run, otherwise an error
            is raised.

        @raises: NoDHCPLeaseError if no leases could be obtained.
        """
        if self.lease:
            return self.lease
        self.lease = maybe_perform_dhcp_discovery(
            self.distro, self.iface, self.dhcp_log_func
        )
        if not self.lease:
            raise NoDHCPLeaseError()
        LOG.debug(
            "Received dhcp lease on %s for %s/%s",
            self.lease["interface"],
            self.lease["fixed-address"],
            self.lease["subnet-mask"],
        )
        nmap = {
            "interface": "interface",
            "ip": "fixed-address",
            "prefix_or_mask": "subnet-mask",
            "broadcast": "broadcast-address",
            "static_routes": [
                "rfc3442-classless-static-routes",
                "classless-static-routes",
                "static_routes",
                "unknown-121",
            ],
            "router": "routers",
        }
        kwargs = self.extract_dhcp_options_mapping(nmap)
        if not kwargs["broadcast"]:
            kwargs["broadcast"] = net.mask_and_ipv4_to_bcast_addr(
                kwargs["prefix_or_mask"], kwargs["ip"]
            )
        if kwargs["static_routes"]:
            kwargs["static_routes"] = (
                self.distro.dhcp_client.parse_static_routes(
                    kwargs["static_routes"]
                )
            )
        ephipv4 = EphemeralIPv4Network(
            self.distro,
            interface_addrs_before_dhcp=self.interface_addrs_before_dhcp,
            **kwargs,
        )
        ephipv4.__enter__()
        self._ephipv4 = ephipv4
        return self.lease

    def extract_dhcp_options_mapping(self, nmap):
        result = {}
        for internal_reference, lease_option_names in nmap.items():
            if isinstance(lease_option_names, list):
                self.get_first_option_value(
                    internal_reference, lease_option_names, result
                )
            else:
                result[internal_reference] = self.lease.get(lease_option_names)
        return result

    def get_first_option_value(
        self, internal_mapping, lease_option_names, result
    ):
        for different_names in lease_option_names:
            if not result.get(internal_mapping):
                result[internal_mapping] = self.lease.get(different_names)


class EphemeralIPNetwork:
    """Combined ephemeral context manager for IPv4 and IPv6

    Either ipv4 or ipv6 ephemeral network may fail to initialize, but if either
    succeeds, then this context manager will not raise exception. This allows
    either ipv4 or ipv6 ephemeral network to succeed, but requires that error
    handling for networks unavailable be done within the context.
    """

    def __init__(
        self,
        distro,
        interface,
        ipv6: bool = False,
        ipv4: bool = True,
        connectivity_urls_data: Optional[List[Dict[str, Any]]] = None,
    ):
        """
        Args:
            distro: The distro object
            interface: The interface to bring up
            ipv6: Whether to bring up an ipv6 network
            ipv4: Whether to bring up an ipv4 network
            connectivity_urls_data: List of url data to use for connectivity
                check before attempting to bring up ephemeral networks. If
                connectivity can be established to any of the urls, then the
                ephemeral network setup is skipped.
        """
        self.interface = interface
        self.ipv4 = ipv4
        self.ipv6 = ipv6
        self.stack = contextlib.ExitStack()
        self.state_msg: str = ""
        self.distro = distro
        self.connectivity_urls_data = connectivity_urls_data

    def __enter__(self):
        if not self.ipv4 and not self.ipv6:
            # no ephemeral network requested, but this object still needs to
            # function as a context manager
            return self
        exceptions = []
        ephemeral_obtained = False

        # short-circuit if we already have connectivity to IMDS
        if imds_url := _check_connectivity_to_imds(
            self.connectivity_urls_data
        ):
            LOG.debug(
                "We already have connectivity to IMDS at %s, skipping DHCP.",
                imds_url,
            )
            return self

        # otherwise, attempt to bring up ephemeral network
        LOG.debug("No connectivity to IMDS, attempting DHCP setup.")

        # first try to bring up ephemeral network for ipv4 (if enabled)
        # then try to bring up ephemeral network for ipv6 (if enabled)
        if self.ipv4:
            ipv4_ephemeral_obtained, ipv4_exception = (
                self._perform_ephemeral_network_setup(ip_version="ipv4")
            )
            ephemeral_obtained |= ipv4_ephemeral_obtained
            if ipv4_exception:
                exceptions.append(ipv4_exception)
        if self.ipv6:
            ipv6_ephemeral_obtained, ipv6_exception = (
                self._perform_ephemeral_network_setup(ip_version="ipv6")
            )
            ephemeral_obtained |= ipv6_ephemeral_obtained
            if ipv6_exception:
                exceptions.append(ipv6_exception)

        # need to set this if we only have ipv6 ephemeral network
        if (self.ipv6 and ipv6_ephemeral_obtained) or not self.ipv4:
            self.state_msg = "using link-local ipv6"

        if not ephemeral_obtained:
            # Ephemeral network setup failed in linkup for both ipv4 and
            # ipv6. Raise only the first exception found.
            LOG.error(
                "Failed to bring up EphemeralIPNetwork. "
                "Datasource setup cannot continue"
            )
            raise exceptions[0]
        return self

    def _perform_ephemeral_network_setup(
        self,
        ip_version: Literal["ipv4", "ipv6"],
    ) -> Tuple[bool, Optional[Exception]]:
        """
        Attempt to bring up an ephemeral network for the specified IP version.

        Args:
            ip_version (str): The IP version to bring up ("ipv4" or "ipv6").

        Returns:
            Tuple: A tuple containing:
                - a boolean indicating whether an ephemeral network was
                    successfully obtained
                - an optional exception if ephemeral network setup failed
                    or None if successful
        """
        try:
            if ip_version == "ipv4":
                self.stack.enter_context(
                    EphemeralDHCPv4(
                        distro=self.distro,
                        iface=self.interface,
                    )
                )
            elif ip_version == "ipv6":
                self.stack.enter_context(
                    EphemeralIPv6Network(
                        self.distro,
                        self.interface,
                    )
                )
            else:
                raise ValueError(f"Unsupported IP version: {ip_version}")

            LOG.debug(
                "Successfully brought up %s for ephemeral %s networking.",
                self.interface,
                ip_version,
            )
            return True, None
        except (ProcessExecutionError, NoDHCPLeaseError) as e:
            LOG.debug(
                "Failed to bring up %s for ephemeral %s networking.",
                self.interface,
                ip_version,
            )
            return False, e

    def __exit__(self, *_args):
        self.stack.close()


def _check_connectivity_to_imds(
    connectivity_urls_data: List[Dict[str, Any]],
) -> Optional[str]:
    """
    Perform a connectivity check to the provided URLs to determine if the
    ephemeral network setup is necessary.

    This function attempts to reach one of the provided URLs and returns the
    URL that was successfully reached. If none of the URLs can be reached,
    it returns None.

    The timeout for the request is determined by the highest timeout value
    provided in the connectivity URLs data. If no timeout is provided, a
    default timeout of 5 seconds is used.

    Args:
        connectivity_urls_data: A list of dictionaries, each containing
            the following keys:
            - "url" (str): The URL to check connectivity for.
            - "headers" (dict, optional): Headers to include in the request.
            - "timeout" (int, optional): Timeout for the request in seconds.

    Returns:
        Optional[str]: The URL that was successfully reached, or None if no
        connectivity was established.
    """

    def _headers_cb(url):
        """
        Helper function to get headers for a given URL from the connectivity
        URLs data provided to _check_connectivity_to_imds.
        """
        headers = [
            url_data.get("headers")
            for url_data in connectivity_urls_data
            if url_data["url"] == url
        ][0]
        return headers

    if not connectivity_urls_data:
        LOG.debug(
            "No connectivity URLs provided. "
            "Skipping connectivity check before ephemeral network setup."
        )
        return None

    # if the user has provided timeout, use the highest value provided
    timeout = (
        max(url_data.get("timeout", 0) for url_data in connectivity_urls_data)
        or 5
    )  # this *should* be sufficient if connectivity is actually available

    try:
        url_that_worked, _ = wait_for_url(
            urls=[url_data["url"] for url_data in connectivity_urls_data],
            headers_cb=_headers_cb,  # get headers per URL
            timeout=timeout,
            connect_synchronously=False,  # use happy eyeballs
            max_wait=0,  # only try once
        )

    # wait_for_url will raise a UrlError if:
    # - request fails
    # - response is empty
    # - response is not OK
    # which is exactly what we want to catch here since if any of these
    # conditions are met, we don't have connectivity to IMDS
    except UrlError as e:
        LOG.debug(
            "Failed to reach IMDS without ephemeral network setup: %s",
            e,
        )
    else:
        # if an error occurs inside wait_for_url that does not result in a
        # UrlError, it won't raise an exception, so we need to check if
        # url_that_worked is None to determine if we have connectivity
        if not url_that_worked:
            LOG.debug("Failed to reach IMDS without ephemeral network setup.")
            return None

        return url_that_worked
    return None



# Copyright (C) 2013 Canonical Ltd.
# Copyright 2019 Joyent, Inc.
#
# Author: Ben Howard <ben.howard@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

# pylint: disable=attribute-defined-outside-init

"""This is a testcase for the SmartOS datasource.

It replicates a serial console and acts like the SmartOS console does in
order to validate return responses.

"""

import json
import multiprocessing
import os
import os.path
import re
import signal
import stat
import uuid
from binascii import crc32
from collections import namedtuple

import pytest
import serial

from cloudinit.atomic_helper import b64e
from cloudinit.event import EventScope, EventType
from cloudinit.sources import DataSourceSmartOS
from cloudinit.sources.DataSourceSmartOS import (
    SERIAL_DEVICE,
    SMARTOS_ENV_KVM,
)
from cloudinit.sources.DataSourceSmartOS import (
    convert_smartos_network_data as convert_net,
)
from cloudinit.sources.DataSourceSmartOS import (
    get_smartos_environ,
    identify_file,
)
from cloudinit.subp import ProcessExecutionError, subp, which
from cloudinit.util import write_file
from tests.unittests.helpers import mock

DSMOS = "cloudinit.sources.DataSourceSmartOS"
SDC_NICS = json.loads(
    """
[
    {
        "nic_tag": "external",
        "primary": true,
        "mtu": 1500,
        "model": "virtio",
        "gateway": "8.12.42.1",
        "netmask": "255.255.255.0",
        "ip": "8.12.42.102",
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "gateways": [
            "8.12.42.1"
        ],
        "vlan_id": 324,
        "mac": "90:b8:d0:f5:e4:f5",
        "interface": "net0",
        "ips": [
            "8.12.42.102/24"
        ]
    },
    {
        "nic_tag": "sdc_overlay/16187209",
        "gateway": "192.168.128.1",
        "model": "virtio",
        "mac": "90:b8:d0:a5:ff:cd",
        "netmask": "255.255.252.0",
        "ip": "192.168.128.93",
        "network_uuid": "4cad71da-09bc-452b-986d-03562a03a0a9",
        "gateways": [
            "192.168.128.1"
        ],
        "vlan_id": 2,
        "mtu": 8500,
        "interface": "net1",
        "ips": [
            "192.168.128.93/22"
        ]
    }
]
"""
)


SDC_NICS_ALT = json.loads(
    """
[
    {
        "interface": "net0",
        "mac": "90:b8:d0:ae:64:51",
        "vlan_id": 324,
        "nic_tag": "external",
        "gateway": "8.12.42.1",
        "gateways": [
          "8.12.42.1"
        ],
        "netmask": "255.255.255.0",
        "ip": "8.12.42.51",
        "ips": [
          "8.12.42.51/24"
        ],
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "model": "virtio",
        "mtu": 1500,
        "primary": true
    },
    {
        "interface": "net1",
        "mac": "90:b8:d0:bd:4f:9c",
        "vlan_id": 600,
        "nic_tag": "internal",
        "netmask": "255.255.255.0",
        "ip": "10.210.1.217",
        "ips": [
          "10.210.1.217/24"
        ],
        "network_uuid": "98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
        "model": "virtio",
        "mtu": 1500
    }
]
"""
)

SDC_NICS_DHCP = json.loads(
    """
[
    {
        "interface": "net0",
        "mac": "90:b8:d0:ae:64:51",
        "vlan_id": 324,
        "nic_tag": "external",
        "gateway": "8.12.42.1",
        "gateways": [
          "8.12.42.1"
        ],
        "netmask": "255.255.255.0",
        "ip": "8.12.42.51",
        "ips": [
          "8.12.42.51/24"
        ],
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "model": "virtio",
        "mtu": 1500,
        "primary": true
    },
    {
        "interface": "net1",
        "mac": "90:b8:d0:bd:4f:9c",
        "vlan_id": 600,
        "nic_tag": "internal",
        "netmask": "255.255.255.0",
        "ip": "10.210.1.217",
        "ips": [
          "dhcp"
        ],
        "network_uuid": "98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
        "model": "virtio",
        "mtu": 1500
    }
]
"""
)

SDC_NICS_MIP = json.loads(
    """
[
    {
        "interface": "net0",
        "mac": "90:b8:d0:ae:64:51",
        "vlan_id": 324,
        "nic_tag": "external",
        "gateway": "8.12.42.1",
        "gateways": [
          "8.12.42.1"
        ],
        "netmask": "255.255.255.0",
        "ip": "8.12.42.51",
        "ips": [
          "8.12.42.51/24",
          "8.12.42.52/24"
        ],
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "model": "virtio",
        "mtu": 1500,
        "primary": true
    },
    {
        "interface": "net1",
        "mac": "90:b8:d0:bd:4f:9c",
        "vlan_id": 600,
        "nic_tag": "internal",
        "netmask": "255.255.255.0",
        "ip": "10.210.1.217",
        "ips": [
          "10.210.1.217/24",
          "10.210.1.151/24"
        ],
        "network_uuid": "98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
        "model": "virtio",
        "mtu": 1500
    }
]
"""
)

SDC_NICS_MIP_IPV6 = json.loads(
    """
[
    {
        "interface": "net0",
        "mac": "90:b8:d0:ae:64:51",
        "vlan_id": 324,
        "nic_tag": "external",
        "gateway": "8.12.42.1",
        "gateways": [
          "8.12.42.1"
        ],
        "netmask": "255.255.255.0",
        "ip": "8.12.42.51",
        "ips": [
          "2001:4800:78ff:1b:be76:4eff:fe06:96b3/64",
          "8.12.42.51/24"
        ],
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "model": "virtio",
        "mtu": 1500,
        "primary": true
    },
    {
        "interface": "net1",
        "mac": "90:b8:d0:bd:4f:9c",
        "vlan_id": 600,
        "nic_tag": "internal",
        "netmask": "255.255.255.0",
        "ip": "10.210.1.217",
        "ips": [
          "10.210.1.217/24"
        ],
        "network_uuid": "98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
        "model": "virtio",
        "mtu": 1500
    }
]
"""
)

SDC_NICS_IPV4_IPV6 = json.loads(
    """
[
    {
        "interface": "net0",
        "mac": "90:b8:d0:ae:64:51",
        "vlan_id": 324,
        "nic_tag": "external",
        "gateway": "8.12.42.1",
        "gateways": ["8.12.42.1", "2001::1", "2001::2"],
        "netmask": "255.255.255.0",
        "ip": "8.12.42.51",
        "ips": ["2001::10/64", "8.12.42.51/24", "2001::11/64",
                "8.12.42.52/32"],
        "network_uuid": "992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
        "model": "virtio",
        "mtu": 1500,
        "primary": true
    },
    {
        "interface": "net1",
        "mac": "90:b8:d0:bd:4f:9c",
        "vlan_id": 600,
        "nic_tag": "internal",
        "netmask": "255.255.255.0",
        "ip": "10.210.1.217",
        "ips": ["10.210.1.217/24"],
        "gateways": ["10.210.1.210"],
        "network_uuid": "98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
        "model": "virtio",
        "mtu": 1500
    }
]
"""
)

SDC_NICS_SINGLE_GATEWAY = json.loads(
    """
[
  {
    "interface":"net0",
    "mac":"90:b8:d0:d8:82:b4",
    "vlan_id":324,
    "nic_tag":"external",
    "gateway":"8.12.42.1",
    "gateways":["8.12.42.1"],
    "netmask":"255.255.255.0",
    "ip":"8.12.42.26",
    "ips":["8.12.42.26/24"],
    "network_uuid":"992fc7ce-6aac-4b74-aed6-7b9d2c6c0bfe",
    "model":"virtio",
    "mtu":1500,
    "primary":true
  },
  {
    "interface":"net1",
    "mac":"90:b8:d0:0a:51:31",
    "vlan_id":600,
    "nic_tag":"internal",
    "netmask":"255.255.255.0",
    "ip":"10.210.1.27",
    "ips":["10.210.1.27/24"],
    "network_uuid":"98657fdf-11f4-4ee2-88a4-ce7fe73e33a6",
    "model":"virtio",
    "mtu":1500
  }
]
"""
)

SDC_NICS_ADDRCONF = json.loads(
    """
[
        {
          "gateway": "10.64.1.129",
          "gateways": [
            "10.64.1.129"
          ],
          "interface": "net0",
          "ip": "10.64.1.130",
          "ips": [
            "10.64.1.130/26",
            "addrconf"
          ],
          "mac": "e2:7f:c1:50:eb:99",
          "model": "virtio",
          "netmask": "255.255.255.192",
          "nic_tag": "external",
          "primary": true,
          "vlan_id": 20
        }
]
"""
)

MOCK_RETURNS = {
    "hostname": "test-host",
    "root_authorized_keys": "ssh-rsa AAAAB3Nz...aC1yc2E= keyname",
    "disable_iptables_flag": None,
    "enable_motd_sys_info": None,
    "test-var1": "some data",
    "cloud-init:user-data": "\n".join(["#!/bin/sh", "/bin/true", ""]),
    "sdc:datacenter_name": "somewhere2",
    "sdc:operator-script": "\n".join(["bin/true", ""]),
    "sdc:uuid": str(uuid.uuid4()),
    "sdc:vendor-data": "\n".join(["VENDOR_DATA", ""]),
    "user-data": "\n".join(["something", ""]),
    "user-script": "\n".join(["/bin/true", ""]),
    "sdc:nics": json.dumps(SDC_NICS),
}

DMI_DATA_RETURN = "smartdc"

# Useful for calculating the length of a frame body.  A SUCCESS body will be
# followed by more characters or be one character less if SUCCESS with no
# payload.  See Section 4.3 of https://eng.joyent.com/mdata/protocol.html.
SUCCESS_LEN = len("0123abcd SUCCESS ")
NOTFOUND_LEN = len("0123abcd NOTFOUND")


class PsuedoJoyentClient:
    def __init__(self, data=None):
        if data is None:
            data = MOCK_RETURNS.copy()
        self.data = data
        self._is_open = False
        return

    def get(self, key, default=None, strip=False):
        if key in self.data:
            r = self.data[key]
            if strip:
                r = r.strip()
        else:
            r = default
        return r

    def get_json(self, key, default=None):
        result = self.get(key, default=default)
        if result is None:
            return default
        return json.loads(result)

    def exists(self):
        return True

    def open_transport(self):
        assert self._is_open is False
        self._is_open = True

    def close_transport(self):
        assert self._is_open
        self._is_open = False


@pytest.fixture
def legacy_user_d(tmp_path):
    legacy_user_dir = str(tmp_path / "legacy_user_tmp")
    os.mkdir(legacy_user_dir)
    return legacy_user_dir


@pytest.fixture
def m_jmc_client_factory(mocker):
    return mocker.patch(
        DSMOS + ".jmc_client_factory",
        return_value=PsuedoJoyentClient(MOCK_RETURNS),
    )


@pytest.fixture
def mocks(legacy_user_d, mocker, m_jmc_client_factory):
    mocker.patch(
        DSMOS + ".get_smartos_environ",
        return_value=DataSourceSmartOS.SMARTOS_ENV_KVM,
    )
    mocker.patch(
        DSMOS + ".LEGACY_USER_D",
        autospec=False,
        new=legacy_user_d,
    )
    mocker.patch(
        DSMOS + ".identify_file",
        return_value="text/plain",
    )
    mocker.patch("cloudinit.net.activators.subp.subp", return_value=("", ""))


def _get_ds(paths, ds_cfg=None):
    sys_cfg = {}
    if ds_cfg is not None:
        sys_cfg["datasource"] = {}
        sys_cfg["datasource"]["SmartOS"] = ds_cfg

    return DataSourceSmartOS.DataSourceSmartOS(
        sys_cfg, distro=None, paths=paths
    )


@pytest.fixture
def ds(paths):
    return lambda: _get_ds(paths)


@pytest.mark.usefixtures("fake_filesystem", "mocks")
class TestSmartOSDataSource:
    jmc_cfact = None
    get_smartos_environ = None

    def test_no_base64(self, paths):
        ds_cfg = {"no_base64_decode": ["test_var1"], "all_base": True}
        dsrc = _get_ds(paths, ds_cfg=ds_cfg)
        ret = dsrc.get_data()
        assert ret is True

    def test_uuid(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["sdc:uuid"] == dsrc.metadata["instance-id"]

    def test_platform_info(self, ds, m_jmc_client_factory):
        """All platform-related attributes are properly set."""
        dsrc = ds()
        assert "joyent" == dsrc.cloud_name
        assert "joyent" == dsrc.platform_type
        assert "serial (/dev/ttyS1)" == dsrc.subplatform

    def test_root_keys(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert (
            MOCK_RETURNS["root_authorized_keys"]
            == dsrc.metadata["public-keys"]
        )

    def test_hostname_b64(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["hostname"] == dsrc.metadata["local-hostname"]

    def test_hostname(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["hostname"] == dsrc.metadata["local-hostname"]

    def test_hostname_if_no_sdc_hostname(self, ds, m_jmc_client_factory):
        my_returns = MOCK_RETURNS.copy()
        my_returns["sdc:hostname"] = "sdc-" + my_returns["hostname"]
        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert my_returns["hostname"] == dsrc.metadata["local-hostname"]

    def test_sdc_hostname_if_no_hostname(self, ds, m_jmc_client_factory):
        my_returns = MOCK_RETURNS.copy()
        my_returns["sdc:hostname"] = "sdc-" + my_returns["hostname"]
        del my_returns["hostname"]
        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert my_returns["sdc:hostname"] == dsrc.metadata["local-hostname"]

    def test_sdc_uuid_if_no_hostname_or_sdc_hostname(
        self, ds, m_jmc_client_factory
    ):
        my_returns = MOCK_RETURNS.copy()
        del my_returns["hostname"]
        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert my_returns["sdc:uuid"] == dsrc.metadata["local-hostname"]

    def test_userdata(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["user-data"] == dsrc.metadata["legacy-user-data"]
        assert MOCK_RETURNS["cloud-init:user-data"] == dsrc.userdata_raw

    def test_sdc_nics(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert (
            json.loads(MOCK_RETURNS["sdc:nics"])
            == dsrc.metadata["network-data"]
        )

    def test_sdc_scripts(self, ds, legacy_user_d, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["user-script"] == dsrc.metadata["user-script"]

        legacy_script_f = "%s/user-script" % legacy_user_d
        print("legacy_script_f=%s" % legacy_script_f)
        assert os.path.exists(legacy_script_f)
        assert os.path.islink(legacy_script_f)
        user_script_perm = oct(os.stat(legacy_script_f)[stat.ST_MODE])[-3:]
        assert user_script_perm == "700"

    def test_scripts_shebanged(self, ds, legacy_user_d, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["user-script"] == dsrc.metadata["user-script"]

        legacy_script_f = "%s/user-script" % legacy_user_d
        assert os.path.exists(legacy_script_f)
        assert os.path.islink(legacy_script_f)
        shebang = None
        with open(legacy_script_f, "r") as f:
            shebang = f.readlines()[0].strip()
        assert shebang == "#!/bin/bash"
        user_script_perm = oct(os.stat(legacy_script_f)[stat.ST_MODE])[-3:]
        assert user_script_perm == "700"

    def test_scripts_shebang_not_added(
        self, ds, legacy_user_d, m_jmc_client_factory
    ):
        """
        Test that the SmartOS requirement that plain text scripts
        are executable. This test makes sure that plain texts scripts
        with out file magic have it added appropriately by cloud-init.
        """

        my_returns = MOCK_RETURNS.copy()
        my_returns["user-script"] = "\n".join(
            ["#!/usr/bin/perl", 'print("hi")', ""]
        )

        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert my_returns["user-script"] == dsrc.metadata["user-script"]

        legacy_script_f = "%s/user-script" % legacy_user_d
        assert os.path.exists(legacy_script_f)
        assert os.path.islink(legacy_script_f)
        shebang = None
        with open(legacy_script_f, "r") as f:
            shebang = f.readlines()[0].strip()
        assert shebang == "#!/usr/bin/perl"

    def test_userdata_removed(self, ds, legacy_user_d, m_jmc_client_factory):
        """
        User-data in the SmartOS world is supposed to be written to a file
        each and every boot. This tests to make sure that in the event the
        legacy user-data is removed, the existing user-data is backed-up
        and there is no /var/db/user-data left.
        """

        user_data_f = "%s/mdata-user-data" % legacy_user_d
        with open(user_data_f, "w") as f:
            f.write("PREVIOUS")

        my_returns = MOCK_RETURNS.copy()
        del my_returns["user-data"]

        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert dsrc.metadata.get("legacy-user-data") is None

        found_new = False
        for root, _dirs, files in os.walk(legacy_user_d):
            for name in files:
                name_f = os.path.join(root, name)
                permissions = oct(os.stat(name_f)[stat.ST_MODE])[-3:]
                if re.match(r".*\/mdata-user-data$", name_f):
                    found_new = True
                    print(name_f)
                    assert permissions == "400"

        assert found_new is False

    def test_vendor_data_not_default(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert MOCK_RETURNS["sdc:vendor-data"] == dsrc.metadata["vendor-data"]

    def test_default_vendor_data(self, ds, m_jmc_client_factory):
        my_returns = MOCK_RETURNS.copy()
        def_op_script = my_returns["sdc:vendor-data"]
        del my_returns["sdc:vendor-data"]
        m_jmc_client_factory.return_value = PsuedoJoyentClient(my_returns)
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert def_op_script != dsrc.metadata["vendor-data"]

        # we expect default vendor-data is a boothook
        assert dsrc.vendordata_raw.startswith("#cloud-boothook")

    def test_disable_iptables_flag(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert (
            MOCK_RETURNS["disable_iptables_flag"]
            == dsrc.metadata["iptables_disable"]
        )

    def test_motd_sys_info(self, ds, m_jmc_client_factory):
        dsrc = ds()
        ret = dsrc.get_data()
        assert ret is True
        assert (
            MOCK_RETURNS["enable_motd_sys_info"]
            == dsrc.metadata["motd_sys_info"]
        )

    def test_default_ephemeral(self, ds):
        # Test to make sure that the builtin config has the ephemeral
        # configuration.
        dsrc = ds()
        cfg = dsrc.get_config_obj()
        ret = dsrc.get_data()
        assert ret is True

        assert "disk_setup" in cfg
        assert "fs_setup" in cfg
        assert isinstance(cfg["disk_setup"], dict)
        assert isinstance(cfg["fs_setup"], list)

    def test_override_disk_aliases(self, paths):
        # Test to make sure that the built-in DS is overridden
        builtin = DataSourceSmartOS.BUILTIN_DS_CONFIG

        mydscfg = {"disk_aliases": {"FOO": "/dev/bar"}}

        # expect that these values are in builtin, or this is pointless
        for k in mydscfg:
            assert k in builtin

        dsrc = _get_ds(paths, ds_cfg=mydscfg)
        ret = dsrc.get_data()
        assert ret is True

        assert (
            mydscfg["disk_aliases"]["FOO"]
            == dsrc.ds_cfg["disk_aliases"]["FOO"]
        )

        assert (
            dsrc.device_name_to_device("FOO") == mydscfg["disk_aliases"]["FOO"]
        )

    def test_reconfig_network_on_boot(self, ds, m_jmc_client_factory):
        # Test to ensure that network is configured from metadata on each boot
        assert {
            EventType.BOOT_NEW_INSTANCE,
            EventType.BOOT,
            EventType.BOOT_LEGACY,
        } == ds().default_update_events[EventScope.NETWORK]


class TestIdentifyFile:
    """Test the 'identify_file' utility."""

    @pytest.mark.allow_subp_for("file")
    @pytest.mark.skipif(
        not which("file"), reason="command 'file' not available."
    )
    def test_file_happy_path(self, tmp_path):
        """Test file is available and functional on plain text."""
        fname = str(tmp_path / "myfile")
        write_file(fname, "plain text content here\n")
        assert "text/plain" == identify_file(fname)

    @mock.patch(DSMOS + ".subp.subp")
    def test_returns_none_on_error(self, m_subp, tmp_path):
        """On 'file' execution error, None should be returned."""
        m_subp.side_effect = ProcessExecutionError("FILE_FAILED", exit_code=99)
        fname = str(tmp_path / "myfile")
        write_file(fname, "plain text content here\n")
        assert None is identify_file(fname)
        assert [
            mock.call(["file", "--brief", "--mime-type", fname])
        ] == m_subp.call_args_list


class ShortReader:
    """Implements a 'read' interface for bytes provided.
    much like io.BytesIO but the 'endbyte' acts as if EOF.
    When it is reached a short will be returned."""

    def __init__(self, initial_bytes, endbyte=b"\0"):
        self.data = initial_bytes
        self.index = 0
        self.len = len(self.data)
        self.endbyte = endbyte

    @property
    def emptied(self):
        return self.index >= self.len

    def read(self, size=-1):
        """Read size bytes but not past a null."""
        if size == 0 or self.index >= self.len:
            return b""

        rsize = size
        if size < 0 or size + self.index > self.len:
            rsize = self.len - self.index

        next_null = self.data.find(self.endbyte, self.index, rsize)
        if next_null >= 0:
            rsize = next_null - self.index + 1
        i = self.index
        self.index += rsize
        ret = self.data[i : i + rsize]
        if len(ret) and ret[-1:] == self.endbyte:
            ret = ret[:-1]
        return ret


@pytest.fixture
def m_serial(mocker):
    return mocker.MagicMock(spec=serial.Serial)


@pytest.fixture
def joyent_metadata(mocker, m_serial):
    res = namedtuple(
        "joyent",
        [
            "serial",
            "request_id",
            "metadata_value",
            "response_parts",
            "metasource_data",
            "metasource_data_len",
        ],
        defaults=(None, None, None, None, None, None),
    )

    res.serial = m_serial
    res.request_id = 0xABCDEF12
    res.metadata_value = "value"
    res.response_parts = {
        "command": "SUCCESS",
        "crc": "b5a9ff00",
        "length": SUCCESS_LEN + len(b64e(res.metadata_value)),
        "payload": b64e(res.metadata_value),
        "request_id": "{0:08x}".format(res.request_id),
    }

    def make_response():
        payloadstr = ""
        if "payload" in res.response_parts:  # pylint: disable=E1135
            payloadstr = " {0}".format(
                res.response_parts["payload"]  # pylint: disable=E1146,E1136
            )
        return (
            "V2 {length} {crc} {request_id} "
            "{command}{payloadstr}\n".format(
                payloadstr=payloadstr,
                **res.response_parts  # pylint: disable=E1134
            ).encode("ascii")
        )

    res.metasource_data = None

    def read_response(length):
        if not res.metasource_data:
            res.metasource_data = make_response()
            res.metasource_data_len = len(res.metasource_data)
        resp = res.metasource_data[:length]  # pylint: disable=E1136
        res.metasource_data = res.metasource_data[  # pylint: disable=E1136
            length:
        ]
        return resp

    res.serial.read.side_effect = read_response
    mocker.patch(
        "cloudinit.sources.DataSourceSmartOS.random.randint",
        mock.Mock(return_value=res.request_id),
    )
    return res


@pytest.fixture
def joyent_client(joyent_metadata):
    return DataSourceSmartOS.JoyentMetadataClient(
        fp=joyent_metadata.serial,
        smartos_type=DataSourceSmartOS.SMARTOS_ENV_KVM,
    )


def _get_written_line(joyent_client, m_serial, key="some_key"):
    joyent_client.get(key)
    return m_serial.write.call_args[0][0]


@pytest.fixture
def joyent_serial_client(joyent_metadata):
    joyent_metadata.serial.timeout = 1
    return DataSourceSmartOS.JoyentMetadataSerialClient(
        None, fp=joyent_metadata.serial
    )


@pytest.mark.usefixtures("fake_filesystem")
class TestJoyentMetadataClient:

    invalid = b"invalid command\n"
    failure = b"FAILURE\n"
    v2_ok = b"V2_OK\n"

    def assertEndsWith(self, haystack, prefix):
        assert haystack.endswith(prefix), "{0} does not end with '{1}'".format(
            repr(haystack), prefix
        )

    def assertStartsWith(self, haystack, prefix):
        assert haystack.startswith(
            prefix
        ), "{0} does not start with '{1}'".format(repr(haystack), prefix)

    def assertNoMoreSideEffects(self, obj):
        with pytest.raises(StopIteration):
            obj()

    def test_get_metadata_writes_a_single_line(self, m_serial, joyent_client):
        joyent_client.get("some_key")
        assert 1 == m_serial.write.call_count
        written_line = m_serial.write.call_args[0][0]
        self.assertEndsWith(
            written_line.decode("ascii"), b"\n".decode("ascii")
        )
        assert 1 == written_line.count(b"\n")

    def test_get_metadata_writes_bytes(self, joyent_client, m_serial):
        assert isinstance(_get_written_line(joyent_client, m_serial), bytes)

    def test_get_metadata_line_starts_with_v2(self, joyent_client, m_serial):
        foo = _get_written_line(joyent_client, m_serial)
        self.assertStartsWith(foo.decode("ascii"), b"V2".decode("ascii"))

    def test_get_metadata_uses_get_command(self, joyent_client, m_serial):
        parts = (
            _get_written_line(joyent_client, m_serial)
            .decode("ascii")
            .strip()
            .split(" ")
        )
        assert "GET" == parts[4]

    def test_get_metadata_base64_encodes_argument(
        self, joyent_client, m_serial
    ):
        key = "my_key"
        parts = (
            _get_written_line(joyent_client, m_serial, key)
            .decode("ascii")
            .strip()
            .split(" ")
        )
        assert b64e(key) == parts[5]

    def test_get_metadata_calculates_length_correctly(
        self, joyent_client, m_serial
    ):
        parts = (
            _get_written_line(joyent_client, m_serial)
            .decode("ascii")
            .strip()
            .split(" ")
        )
        expected_length = len(" ".join(parts[3:]))
        assert expected_length == int(parts[1])

    def test_get_metadata_uses_appropriate_request_id(
        self, joyent_client, m_serial
    ):
        parts = (
            _get_written_line(joyent_client, m_serial)
            .decode("ascii")
            .strip()
            .split(" ")
        )
        request_id = parts[3]
        assert 8 == len(request_id)
        assert request_id == request_id.lower()

    def test_get_metadata_uses_random_number_for_request_id(
        self, joyent_client, joyent_metadata, m_serial
    ):
        line = _get_written_line(joyent_client, m_serial)
        request_id = line.decode("ascii").strip().split(" ")[3]
        assert "{0:08x}".format(joyent_metadata.request_id) == request_id

    def test_get_metadata_checksums_correctly(self, joyent_client, m_serial):
        parts = (
            _get_written_line(joyent_client, m_serial)
            .decode("ascii")
            .strip()
            .split(" ")
        )
        expected_checksum = "{0:08x}".format(
            crc32(" ".join(parts[3:]).encode("utf-8")) & 0xFFFFFFFF
        )
        checksum = parts[2]
        assert expected_checksum == checksum

    def test_get_metadata_reads_a_line(
        self, joyent_client, joyent_metadata, m_serial
    ):
        joyent_client.get("some_key")
        assert joyent_metadata.metasource_data_len == m_serial.read.call_count

    def test_get_metadata_returns_valid_value(
        self, joyent_client, joyent_metadata
    ):
        value = joyent_client.get("some_key")
        assert joyent_metadata.metadata_value == value

    def test_get_metadata_throws_exception_for_incorrect_length(
        self, joyent_client, joyent_metadata
    ):
        joyent_metadata.response_parts["length"] = 0
        with pytest.raises(DataSourceSmartOS.JoyentMetadataFetchException):
            joyent_client.get(
                "some_key",
            )

    def test_get_metadata_throws_exception_for_incorrect_crc(
        self, joyent_client, joyent_metadata
    ):
        joyent_metadata.response_parts["crc"] = "deadbeef"
        with pytest.raises(DataSourceSmartOS.JoyentMetadataFetchException):
            joyent_client.get(
                "some_key",
            )

    def test_get_metadata_throws_exception_for_request_id_mismatch(
        self, joyent_client, joyent_metadata
    ):
        joyent_metadata.response_parts["request_id"] = "deadbeef"
        joyent_client._checksum = lambda _: joyent_metadata.response_parts[
            "crc"
        ]
        with pytest.raises(DataSourceSmartOS.JoyentMetadataFetchException):
            joyent_client.get("some_key")

    def test_get_metadata_returns_None_if_value_not_found(
        self, joyent_client, joyent_metadata
    ):
        joyent_metadata.response_parts["payload"] = ""
        joyent_metadata.response_parts["command"] = "NOTFOUND"
        joyent_metadata.response_parts["length"] = NOTFOUND_LEN
        joyent_client._checksum = lambda _: joyent_metadata.response_parts[
            "crc"
        ]
        assert joyent_client.get("some_key") is None

    def test_negotiate(self, joyent_client):
        reader = ShortReader(self.v2_ok)
        joyent_client.fp.read.side_effect = reader.read
        joyent_client._negotiate()
        assert reader.emptied

    def test_negotiate_short_response(self, joyent_client):
        # chopped '\n' from v2_ok.
        reader = ShortReader(self.v2_ok[:-1] + b"\0")
        joyent_client.fp.read.side_effect = reader.read
        with pytest.raises(DataSourceSmartOS.JoyentMetadataTimeoutException):
            joyent_client._negotiate()
        assert reader.emptied

    def test_negotiate_bad_response(self, joyent_client):
        reader = ShortReader(b"garbage\n" + self.v2_ok)
        joyent_client.fp.read.side_effect = reader.read
        with pytest.raises(DataSourceSmartOS.JoyentMetadataFetchException):
            joyent_client._negotiate()
        assert self.v2_ok == joyent_client.fp.read()

    def test_serial_open_transport(self, joyent_serial_client):
        reader = ShortReader(b"garbage\0" + self.invalid + self.v2_ok)
        joyent_serial_client.fp.read.side_effect = reader.read
        joyent_serial_client.open_transport()
        assert reader.emptied

    def test_flush_failure(self, joyent_serial_client):
        reader = ShortReader(
            b"garbage" + b"\0" + self.failure + self.invalid + self.v2_ok
        )
        joyent_serial_client.fp.read.side_effect = reader.read
        joyent_serial_client.open_transport()
        assert reader.emptied

    def test_flush_many_timeouts(self, joyent_serial_client):
        reader = ShortReader(b"\0" * 100 + self.invalid + self.v2_ok)
        joyent_serial_client.fp.read.side_effect = reader.read
        joyent_serial_client.open_transport()
        assert reader.emptied

    def test_list_metadata_returns_list(self, joyent_client, joyent_metadata):
        parts = ["foo", "bar"]
        value = b64e("\n".join(parts))
        joyent_metadata.response_parts["payload"] = value
        joyent_metadata.response_parts["crc"] = "40873553"
        joyent_metadata.response_parts["length"] = SUCCESS_LEN + len(value)
        assert joyent_client.list() == parts

    def test_list_metadata_returns_empty_list_if_no_customer_metadata(
        self, joyent_client, joyent_metadata
    ):
        del joyent_metadata.response_parts["payload"]
        joyent_metadata.response_parts["length"] = SUCCESS_LEN - 1
        joyent_metadata.response_parts["crc"] = "14e563ba"
        assert joyent_client.list() == []


class TestNetworkConversion:
    def test_convert_simple(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.102/24",
                        }
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:f5:e4:f5",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"type": "static", "address": "192.168.128.93/22"}
                    ],
                    "mtu": 8500,
                    "mac_address": "90:b8:d0:a5:ff:cd",
                },
            ],
        }
        found = convert_net(SDC_NICS)
        assert expected == found

    def test_convert_simple_alt(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.51/24",
                        }
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:ae:64:51",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"type": "static", "address": "10.210.1.217/24"}
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:bd:4f:9c",
                },
            ],
        }
        found = convert_net(SDC_NICS_ALT)
        assert expected == found

    def test_convert_simple_dhcp(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.51/24",
                        }
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:ae:64:51",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [{"type": "dhcp4"}],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:bd:4f:9c",
                },
            ],
        }
        found = convert_net(SDC_NICS_DHCP)
        assert expected == found

    def test_convert_simple_multi_ip(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.51/24",
                        },
                        {"type": "static", "address": "8.12.42.52/24"},
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:ae:64:51",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"type": "static", "address": "10.210.1.217/24"},
                        {"type": "static", "address": "10.210.1.151/24"},
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:bd:4f:9c",
                },
            ],
        }
        found = convert_net(SDC_NICS_MIP)
        assert expected == found

    def test_convert_with_dns(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.51/24",
                        }
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:ae:64:51",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [{"type": "dhcp4"}],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:bd:4f:9c",
                },
                {
                    "type": "nameserver",
                    "address": ["8.8.8.8", "8.8.8.1"],
                    "search": ["local"],
                },
            ],
        }
        found = convert_net(
            network_data=SDC_NICS_DHCP,
            dns_servers=["8.8.8.8", "8.8.8.1"],
            dns_domain="local",
        )
        assert expected == found

    def test_convert_simple_multi_ipv6(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "type": "static",
                            "address": (
                                "2001:4800:78ff:1b:be76:4eff:fe06:96b3/64"
                            ),
                        },
                        {
                            "type": "static",
                            "gateway": "8.12.42.1",
                            "address": "8.12.42.51/24",
                        },
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:ae:64:51",
                },
                {
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"type": "static", "address": "10.210.1.217/24"}
                    ],
                    "mtu": 1500,
                    "mac_address": "90:b8:d0:bd:4f:9c",
                },
            ],
        }
        found = convert_net(SDC_NICS_MIP_IPV6)
        assert expected == found

    def test_convert_simple_both_ipv4_ipv6(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "mac_address": "90:b8:d0:ae:64:51",
                    "mtu": 1500,
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "address": "2001::10/64",
                            "gateway": "2001::1",
                            "type": "static",
                        },
                        {
                            "address": "8.12.42.51/24",
                            "gateway": "8.12.42.1",
                            "type": "static",
                        },
                        {"address": "2001::11/64", "type": "static"},
                        {"address": "8.12.42.52/32", "type": "static"},
                    ],
                },
                {
                    "mac_address": "90:b8:d0:bd:4f:9c",
                    "mtu": 1500,
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"address": "10.210.1.217/24", "type": "static"}
                    ],
                },
            ],
        }
        found = convert_net(SDC_NICS_IPV4_IPV6)
        assert expected == found

    def test_gateways_not_on_all_nics(self):
        expected = {
            "version": 1,
            "config": [
                {
                    "mac_address": "90:b8:d0:d8:82:b4",
                    "mtu": 1500,
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "address": "8.12.42.26/24",
                            "gateway": "8.12.42.1",
                            "type": "static",
                        }
                    ],
                },
                {
                    "mac_address": "90:b8:d0:0a:51:31",
                    "mtu": 1500,
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {"address": "10.210.1.27/24", "type": "static"}
                    ],
                },
            ],
        }
        found = convert_net(SDC_NICS_SINGLE_GATEWAY)
        assert expected == found

    def test_routes_on_all_nics(self):
        routes = [
            {"linklocal": False, "dst": "3.0.0.0/8", "gateway": "8.12.42.3"},
            {"linklocal": False, "dst": "4.0.0.0/8", "gateway": "10.210.1.4"},
        ]
        expected = {
            "version": 1,
            "config": [
                {
                    "mac_address": "90:b8:d0:d8:82:b4",
                    "mtu": 1500,
                    "name": "net0",
                    "type": "physical",
                    "subnets": [
                        {
                            "address": "8.12.42.26/24",
                            "gateway": "8.12.42.1",
                            "type": "static",
                            "routes": [
                                {
                                    "network": "3.0.0.0/8",
                                    "gateway": "8.12.42.3",
                                },
                                {
                                    "network": "4.0.0.0/8",
                                    "gateway": "10.210.1.4",
                                },
                            ],
                        }
                    ],
                },
                {
                    "mac_address": "90:b8:d0:0a:51:31",
                    "mtu": 1500,
                    "name": "net1",
                    "type": "physical",
                    "subnets": [
                        {
                            "address": "10.210.1.27/24",
                            "type": "static",
                            "routes": [
                                {
                                    "network": "3.0.0.0/8",
                                    "gateway": "8.12.42.3",
                                },
                                {
                                    "network": "4.0.0.0/8",
                                    "gateway": "10.210.1.4",
                                },
                            ],
                        }
                    ],
                },
            ],
        }
        found = convert_net(SDC_NICS_SINGLE_GATEWAY, routes=routes)
        self.maxDiff = None
        assert expected == found

    def test_ipv6_addrconf(self):
        expected = {
            "config": [
                {
                    "mac_address": "e2:7f:c1:50:eb:99",
                    "name": "net0",
                    "subnets": [
                        {
                            "address": "10.64.1.130/26",
                            "gateway": "10.64.1.129",
                            "type": "static",
                        },
                        {"type": "dhcp6"},
                    ],
                    "type": "physical",
                }
            ],
            "version": 1,
        }
        found = convert_net(SDC_NICS_ADDRCONF)
        self.maxDiff = None
        assert expected == found


@pytest.fixture
def mdata_proc():
    mdata_proc = multiprocessing.Process(target=start_mdata_loop)
    mdata_proc.start()

    yield mdata_proc

    # os.kill() rather than mdata_proc.terminate() to avoid console spam.
    os.kill(mdata_proc.pid, signal.SIGKILL)
    mdata_proc.join()


def start_mdata_loop():
    """
    The mdata-get command is repeatedly run in a separate process so
    that it may try to race with metadata operations performed in the
    main test process.  Use of mdata-get is better than two processes
    using the protocol implementation in DataSourceSmartOS because we
    are testing to be sure that cloud-init and mdata-get respect each
    others locks.
    """
    rcs = list(range(256))
    while True:
        subp(["mdata-get", "sdc:routes"], rcs=rcs)


@pytest.mark.skipif(
    get_smartos_environ() != SMARTOS_ENV_KVM,
    reason="Only supported on KVM and bhyve guests under SmartOS",
)
@pytest.mark.skipif(
    not os.access(SERIAL_DEVICE, os.W_OK),
    reason="Requires write access to " + SERIAL_DEVICE,
)
class TestSerialConcurrency:
    """
    This class tests locking on an actual serial port, and as such can only
    be run in a kvm or bhyve guest running on a SmartOS host.  A test run on
    a metadata socket will not be valid because a metadata socket ensures
    there is only one session over a connection.  In contrast, in the
    absence of proper locking multiple processes opening the same serial
    port can corrupt each others' exchanges with the metadata server.

    This takes on the order of 2 to 3 minutes to run.
    """

    @pytest.mark.allow_subp_for("mdata-get")
    def test_all_keys(self, mdata_proc):
        assert mdata_proc.pid is not None
        ds = DataSourceSmartOS
        keys = [tup[0] for tup in ds.SMARTOS_ATTRIB_MAP.values()]
        keys.extend(ds.SMARTOS_ATTRIB_JSON.values())

        client = ds.jmc_client_factory(smartos_type=SMARTOS_ENV_KVM)
        assert client is not None

        # The behavior that we are testing for was observed mdata-get running
        # 10 times at roughly the same time as cloud-init fetched each key
        # once.  cloud-init would regularly see failures before making it
        # through all keys once.
        for _ in range(3):
            for key in keys:
                # We don't care about the return value, just that it doesn't
                # thrown any exceptions.
                client.get(key)

        assert mdata_proc.exitcode is None



# Copyright (C) 2024 Canonical Ltd.
#
# Author: Carlos Nihelton <carlos.santanadeoliveira@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
import logging
import os
import re
from copy import deepcopy
from email.mime.multipart import MIMEMultipart
from pathlib import PurePath
from typing import cast

import pytest

from cloudinit import util
from cloudinit.sources import DataSourceWSL as wsl
from tests.unittests.helpers import does_not_raise, get_distro, mock

INSTANCE_NAME = "Noble-MLKit"
GOOD_MOUNTS = {
    "none": {
        "fstype": "tmpfs",
        "mountpoint": "/mnt/wsl",
        "opts": "rw,relatime",
    },
    "/dev/sdd": {
        "fstype": "ext4",
        "mountpoint": "/",
        "opts": "rw,relatime,...",
    },
    "sysfs": {"fstype": "sysfs", "mountpoint": "/sys", "opts": "rw,nosuid..."},
    "C:\\": {
        "fstype": "9p",
        "mountpoint": "/mnt/c",
        "opts": "rw,noatime,dirsync,aname=drvfs;path=C:\\;...",
    },
    "D:\\": {
        "fstype": "9p",
        "mountpoint": "/mnt/d",
        "opts": "rw,noatime,dirsync,aname=drvfs;path=D:\\;...",
    },
    "hugetblfs": {
        "fstype": "hugetblfs",
        "mountpoint": "/dev/hugepages",
        "opts": "rw,relatime...",
    },
}
SAMPLE_LINUX_DISTRO = ("ubuntu", "24.04", "noble")
SAMPLE_LINUX_DISTRO_NO_VERSION_ID = ("debian", "", "trixie")

AGENT_SAMPLE = """\
#cloud-config
landscape:
    host:
        url: landscape.canonical.com:6554
    client:
        account_name: agenttest
        url: https://landscape.canonical.com/message-system
        ping_url: https://landscape.canonical.com/ping
        tags: wsl
ubuntu_pro:
    token: testtoken
"""

LANDSCAPE_SAMPLE = """\
#cloud-config
landscape:
  client:
    account_name: landscapetest
    tags: tag_aiml,tag_dev
locale: en_GB.UTF-8
"""


class TestWSLHelperFunctions:
    @mock.patch("cloudinit.util.subp.subp")
    def test_instance_name(self, m_subp):
        m_subp.return_value = util.subp.SubpResult(
            f"//wsl.localhost/{INSTANCE_NAME}/", ""
        )

        assert INSTANCE_NAME == wsl.instance_name()

    @mock.patch("cloudinit.util.mounts")
    def test_mounted_drives(self, m_mounts):
        # A good output
        m_mounts.return_value = deepcopy(GOOD_MOUNTS)
        mounts = wsl.mounted_win_drives()
        assert ["/mnt/c", "/mnt/d"] == mounts

        # no more drvfs in C:\ options
        m_mounts.return_value["C:\\"]["opts"] = "rw,relatime..."
        mounts = wsl.mounted_win_drives()
        assert ["/mnt/d"] == mounts

        # fstype mismatch for D:\
        m_mounts.return_value["D:\\"]["fstype"] = "zfs"
        mounts = wsl.mounted_win_drives()
        assert [] == mounts

    @mock.patch("os.access")
    @mock.patch("cloudinit.util.mounts")
    def test_cmd_exe_ok(self, m_mounts, m_os_access):
        """
        Validates the happy path, when we find the Windows system drive and
        cmd.exe is executable.
        """
        m_mounts.return_value = deepcopy(GOOD_MOUNTS)
        m_os_access.return_value = True
        cmd = wsl.cmd_executable()
        # To please pyright not to complain about optional member access.
        assert cmd is not None
        assert None is not cmd.relative_to(GOOD_MOUNTS["C:\\"]["mountpoint"])

    @mock.patch("os.access")
    @mock.patch("cloudinit.util.mounts")
    def test_cmd_not_executable(self, m_mounts, m_os_access):
        """
        When the cmd.exe found is not executable, then RuntimeError is raised.
        """
        m_mounts.return_value = deepcopy(GOOD_MOUNTS)
        m_os_access.return_value = True
        cmd = wsl.cmd_executable()
        # To please pyright not to complain about optional member access.
        assert cmd is not None
        assert None is not cmd.relative_to(GOOD_MOUNTS["C:\\"]["mountpoint"])

        m_os_access.return_value = False
        with pytest.raises(IOError):
            wsl.cmd_executable()

    @mock.patch("os.access")
    @mock.patch("cloudinit.util.mounts")
    def test_cmd_exe_no_win_mounts(self, m_mounts, m_os_access):
        """
        When no Windows drives are found, then RuntimeError is raised.
        """
        m_os_access.return_value = True

        m_mounts.return_value = deepcopy(GOOD_MOUNTS)
        m_mounts.return_value.pop("C:\\")
        m_mounts.return_value.pop("D:\\")
        with pytest.raises(IOError):
            wsl.cmd_executable()

    @pytest.mark.parametrize(
        "linux_distro_value,files",
        (
            (
                SAMPLE_LINUX_DISTRO,
                [
                    f"{INSTANCE_NAME}.user-data",
                    "ubuntu-24.04.user-data",
                    "ubuntu-all.user-data",
                    "default.user-data",
                ],
            ),
            (
                SAMPLE_LINUX_DISTRO_NO_VERSION_ID,
                [
                    f"{INSTANCE_NAME}.user-data",
                    "debian-trixie.user-data",
                    "debian-all.user-data",
                    "default.user-data",
                ],
            ),
        ),
    )
    @mock.patch("cloudinit.util.get_linux_distro")
    def test_candidate_files(self, m_gld, linux_distro_value, files):
        """
        Validate the file names candidate for holding user-data and their
        order of precedence.
        """
        m_gld.return_value = linux_distro_value
        assert files == wsl.candidate_user_data_file_names(INSTANCE_NAME)

    @pytest.mark.parametrize(
        "md_content,is_from_pro,raises,errors,warnings,md_expected",
        (
            pytest.param(
                None,
                False,
                does_not_raise(),
                [],
                [],
                {"instance-id": "iid-datasource-wsl"},
                id="default_md_on_no_md_file",
            ),
            pytest.param(
                '{"instance-id":"iid-load-from-pro"}',
                True,
                does_not_raise(),
                [],
                [],
                {"instance-id": "iid-load-from-pro"},
                id="metadata_from_pro",
            ),
            pytest.param(
                "{}",
                False,
                pytest.raises(
                    ValueError,
                    match=(
                        "myinstance.meta-data does not contain instance-id key"
                    ),
                ),
                ["myinstance.meta-data does not contain instance-id key"],
                [],
                "",
                id="error_on_md_missing_instance_id_key",
            ),
            pytest.param(
                "{",
                True,
                pytest.raises(
                    ValueError,
                    match=(
                        "myinstance.meta-data does not contain instance-id key"
                    ),
                ),
                ["myinstance.meta-data does not contain instance-id key"],
                ["Failed loading yaml blob. Invalid format at line 1"],
                "",
                id="error_on_md_invalid_yaml",
            ),
        ),
    )
    def test_load_instance_metadata(
        self,
        md_content,
        is_from_pro,
        raises,
        errors,
        warnings,
        md_expected,
        tmpdir,
        caplog,
    ):
        """meta-data file is optional. Errors are raised on invalid content."""
        path = ".cloud-init"
        if is_from_pro:
            path = ".ubuntupro/.cloud-init"

        if md_content is not None:
            dir = tmpdir.join(path)
            os.makedirs(dir)
            dir.join("myinstance.meta-data").write(md_content)
        with caplog.at_level(logging.WARNING):
            with raises:
                assert md_expected == wsl.load_instance_metadata(
                    tmpdir, "myinstance"
                )
            warning_logs = "\n".join(
                [
                    x.message
                    for x in caplog.records
                    if x.levelno == logging.WARNING
                ]
            )
            error_logs = "\n".join(
                [
                    x.message
                    for x in caplog.records
                    if x.levelno == logging.ERROR
                ]
            )
        if warnings:
            for warning in warnings:
                assert warning in warning_logs
        else:
            assert "" == warning_logs
        if errors:
            for error in errors:
                assert error in error_logs
        else:
            assert "" == error_logs

    @mock.patch("cloudinit.util.subp.subp")
    def test_landscape_supports_field(self, m_subp):
        LANDSCAPE_HELP_OUTPUT = """\
        Usage: landscape-config [options]

        Options:
          --version             show program's version number and exit
          -h, --help            show this help message and exit
          --installation-request-id Only set this value if this computer is a
                                    instance managed by Landscape, in which
                                    it to be the request id that Landscape
                                    to the installation activity for the host
          --access-group=ACCESS_GROUP
                                Suggested access group for this computer.
          --tags=TAGS           Comma separated list of tag names to be sent
                                to the server.
        """
        m_subp.return_value = util.subp.SubpResult(LANDSCAPE_HELP_OUTPUT, "")
        assert wsl.landscape_supports_field("non_sense") is False
        assert wsl.landscape_supports_field("tags") is True
        assert wsl.landscape_supports_field("installation-request-id") is True


SAMPLE_CFG = {"datasource_list": ["NoCloud", "WSL"]}


def join_payloads_from_content_type(
    part: MIMEMultipart, content_type: str
) -> str:
    """
    Helper function to decode and join all parts of a multipart MIME
    message matched by the content type.
    """
    content = ""
    for p in part.walk():
        if p.get_content_type() == content_type:
            content = content + str(p.get_payload())

    return content


class TestMergeAgentLandscapeData:
    @pytest.mark.parametrize(
        "agent_yaml,landscape_user_data,expected",
        (
            pytest.param(
                None, None, None, id="none_when_both_agent_and_ud_none"
            ),
            pytest.param(
                None, "", None, id="none_when_agent_none_and_ud_empty"
            ),
            pytest.param(
                "", None, None, id="none_when_agent_empty_and_ud_none"
            ),
            pytest.param("", "", None, id="none_when_both_agent_and_ud_empty"),
            pytest.param(
                AGENT_SAMPLE, "", AGENT_SAMPLE, id="agent_only_when_ud_empty"
            ),
            pytest.param(
                "",
                LANDSCAPE_SAMPLE,
                LANDSCAPE_SAMPLE,
                id="ud_only_when_agent_empty",
            ),
            pytest.param(
                "#cloud-config\nlandscape:\n client: {account_name: agent}\n",
                LANDSCAPE_SAMPLE,
                "#cloud-config\n# WSL datasouce Merged agent.yaml and "
                "user_data\n"
                + "\n".join(LANDSCAPE_SAMPLE.splitlines()[1:]).replace(
                    "landscapetest", "agent"
                ),
                id="merge_agent_and_landscape_ud_when_both_present",
            ),
        ),
    )
    def test_merged_data_excludes_empty_or_none(
        self, agent_yaml, landscape_user_data, expected, tmpdir
    ):
        agent_data = user_data = None
        if agent_yaml is not None:
            agent_path = tmpdir.join("agent.yaml")
            agent_path.write(agent_yaml)
            agent_data = wsl.ConfigData(agent_path, "")
        if landscape_user_data is not None:
            landscape_ud_path = tmpdir.join("instance_name.user_data")
            landscape_ud_path.write(landscape_user_data)
            user_data = wsl.ConfigData(landscape_ud_path, "")
        assert expected == wsl.merge_agent_landscape_data(
            agent_data, user_data
        )


class TestWSLDataSource:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, tmpdir):
        mocker.patch(
            "cloudinit.sources.DataSourceWSL.instance_name",
            return_value=INSTANCE_NAME,
        )
        mocker.patch(
            "cloudinit.sources.DataSourceWSL.find_home",
            return_value=PurePath(tmpdir),
        )
        mocker.patch(
            "cloudinit.sources.DataSourceWSL.subp.which",
            return_value="/usr/bin/wslpath",
        )
        mocker.patch(
            "cloudinit.sources.DataSourceWSL.landscape_supports_field",
            return_value=True,
        )

    def test_metadata_id_default(self, tmpdir, paths):
        """
        Validates that instance-id is properly set, independent of the
        existence of user-data.
        """

        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )
        ds.get_data()

        assert ds.get_instance_id() == wsl.DEFAULT_INSTANCE_ID

    def test_metadata_id(self, tmpdir, paths):
        """
        Validates that instance-id is properly set, independent of the
        existence of user-data.
        """
        SAMPLE_ID = "Nice-ID"
        metadata_path = tmpdir.join(
            ".cloud-init", f"{INSTANCE_NAME}.meta-data"
        )
        metadata_path.dirpath().mkdir()
        metadata_path.write(f'{{"instance-id":"{SAMPLE_ID}"}}')

        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )
        ds.get_data()

        assert ds.get_instance_id() == SAMPLE_ID

    @mock.patch("cloudinit.util.lsb_release")
    def test_get_data_cc(self, m_lsb_release, paths, tmpdir):
        m_lsb_release.return_value = SAMPLE_LINUX_DISTRO
        data_path = tmpdir.join(".cloud-init", f"{INSTANCE_NAME}.user-data")
        data_path.dirpath().mkdir()
        data_path.write("#cloud-config\nwrite_files:\n- path: /etc/wsl.conf")

        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )
        assert userdata is not None
        assert "wsl.conf" in userdata

    @mock.patch("cloudinit.util.lsb_release")
    def test_get_data_sh(self, m_lsb_release, tmpdir, paths):
        m_lsb_release.return_value = SAMPLE_LINUX_DISTRO
        COMMAND = "echo Hello cloud-init on WSL!"
        data_path = tmpdir.join(".cloud-init", f"{INSTANCE_NAME}.user-data")
        data_path.dirpath().mkdir()
        data_path.write(f"#!/bin/sh\n{COMMAND}\n")
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/x-shellscript"
        )
        assert COMMAND in userdata

    @mock.patch("cloudinit.util.lsb_release")
    def test_get_data_jinja(self, m_lsb_release, paths, tmpdir):
        """Assert we don't mistakenly treat jinja as final cloud-config"""
        m_lsb_release.return_value = SAMPLE_LINUX_DISTRO
        data_path = tmpdir.join(".cloud-init", f"{INSTANCE_NAME}.user-data")
        data_path.dirpath().mkdir()
        data_path.write(
            """## template: jinja
#cloud-config
write_files:
- path: /etc/{{ v1.instance_name }}.conf
"""
        )

        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata(True)
        print(ud)

        assert ud is not None
        assert "write_files" in join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/jinja2"
        ), "Jinja should not be treated as final cloud-config"
        assert "write_files" not in join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        ), "No cloud-config part should exist"

    @pytest.mark.parametrize("with_agent_data", [True, False])
    @mock.patch("cloudinit.util.lsb_release")
    def test_get_data_x(
        self, m_lsb_release, with_agent_data, caplog, paths, tmpdir
    ):
        """
        Assert behavior of empty .cloud-config dir with and without agent data
        """
        m_lsb_release.return_value = SAMPLE_LINUX_DISTRO
        data_path = tmpdir.join(".cloud-init", f"{INSTANCE_NAME}.user-data")
        data_path.dirpath().mkdir()

        if with_agent_data:
            ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
            os.makedirs(ubuntu_pro_tmp, exist_ok=True)
            agent_path = ubuntu_pro_tmp.join("agent.yaml")
            agent_path.write(AGENT_SAMPLE)

        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is with_agent_data
        if with_agent_data:
            assert ds.userdata_raw == AGENT_SAMPLE
        else:
            assert ds.userdata_raw is None

        expected_log_level = logging.INFO if with_agent_data else logging.ERROR
        regex = (
            "Unable to load any user-data file in /[^:]*/.cloud-init:"
            " /.*/.cloud-init directory is empty"
        )
        messages = [
            x.message
            for x in caplog.records
            if x.levelno == expected_log_level and re.match(regex, x.message)
        ]
        assert (
            len(messages) > 0
        ), "Expected log message matching '{}' with log level '{}'".format(
            regex, expected_log_level
        )

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_data_precedence(self, m_get_linux_dist, tmpdir, paths):
        """Validates the precedence of user-data files."""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        # Set up basic user data:

        # This is the most specific: should win over the other user-data files.
        # Also, notice the file name casing: should be irrelevant.
        user_file = tmpdir.join(".cloud-init", "ubuntu-24.04.user-data")
        user_file.dirpath().mkdir()
        user_file.write("#cloud-config\nwrite_files:\n- path: /etc/wsl.conf")

        distro_file = tmpdir.join(".cloud-init", "Ubuntu-all.user-data")
        distro_file.write("#!/bin/sh\n\necho Hello World\n")

        generic_file = tmpdir.join(".cloud-init", "default.user-data")
        generic_file.write("#cloud-config\npackages:\n- g++-13\n")

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        # Assert user data is properly loaded
        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )
        assert "wsl.conf" in userdata
        assert "packages" not in userdata
        shell_script = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/x-shellscript"
        )

        assert "" == shell_script

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_interaction_with_pro(self, m_get_linux_dist, tmpdir, paths):
        """Validates the interaction of user-data and Pro For WSL agent data"""

        m_get_linux_dist.return_value = ("ubuntu", "25.10", "plucky")

        user_file = tmpdir.join(".cloud-init", "ubuntu-25.10.user-data")
        user_file.dirpath().mkdir()
        user_file.write("#cloud-config\nwrite_files:\n- path: /etc/wsl.conf")

        # The winner should be the merge of the agent and user provided data.
        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        agent_file.write(
            """#cloud-config
landscape:
    host:
        url: landscape.canonical.com:6554
    client:
        account_name: agenttest
        url: https://landscape.canonical.com/message-system
        ping_url: https://landscape.canonical.com/ping
        tags: wsl
ubuntu_pro:
    token: testtoken"""
        )
        SAMPLE_ID = "Nice-ID"
        agent_metadata_path = ubuntu_pro_tmp.join(f"{INSTANCE_NAME}.meta-data")
        agent_metadata_path.write(f'{{"instance-id":"{SAMPLE_ID}"}}')

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        # Assert agent combines with existing user data
        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )
        assert "wsl.conf" in userdata
        assert "packages" not in userdata
        assert "ubuntu_pro" in userdata
        assert "landscape" in userdata
        assert "agenttest" in userdata
        assert "installation_request_id" in userdata
        assert SAMPLE_ID in userdata

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_landscape_vs_local_user(self, m_get_linux_dist, tmpdir, paths):
        """Validates the precedence of Landscape-provided over local data"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        user_file = tmpdir.join(".cloud-init", "ubuntu-24.04.user-data")
        user_file.dirpath().mkdir()
        user_file.write(
            """#cloud-config
ubuntu_pro:
    token: usertoken
package_update: true"""
        )

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)
        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(LANDSCAPE_SAMPLE)

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()
        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )

        assert (
            "locale" in userdata
            and "landscapetest" in userdata
            and "ubuntu_pro" not in userdata
            and "package_update" not in userdata
        ), "Landscape data should have overridden user provided data"

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_landscape_provided_data(self, m_get_linux_dist, tmpdir, paths):
        """Validates the interaction of Pro For WSL agent and Landscape data"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        agent_file.write(
            """#cloud-config
landscape:
    host:
        url: hosted.com:6554
    client:
        account_name: agenttest
        url: https://hosted.com/message-system
        ping_url: https://hosted.com/ping
        ssl_public_key: C:\\Users\\User\\server.pem
        tags: wsl
ubuntu_pro:
    token: testtoken"""
        )

        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(
            """#cloud-config
landscape:
  client:
    account_name: landscapetest
    tags: tag_aiml,tag_dev
package_update: true"""
        )

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        # Assert Landscape and Agent combine, with Agent taking precedence
        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )

        assert "ubuntu_pro" in userdata, "Agent data should be present"
        assert "package_update" in userdata, (
            "package_update entry should not be overridden by agent data"
            " nor ignored"
        )
        assert (
            "landscapetest" not in userdata and "agenttest" in userdata
        ), "Landscape account name should have been overridden by agent data"
        # Make sure we have tags from Landscape data, not agent's
        assert (
            "tag_aiml" in userdata and "tag_dev" in userdata
        ), "User-data should override agent data's Landscape computer tags"
        assert "wsl" not in userdata

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_landscape_empty_data(self, m_get_linux_dist, tmpdir, paths):
        """Asserts that Pro for WSL data is present when Landscape is empty"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        agent_file.write(
            """#cloud-config
landscape:
    host:
        url: hosted.com:6554
    client:
        account_name: agent_test
        url: https://hosted.com/message-system
        ping_url: https://hosted.com/ping
        ssl_public_key: C:\\Users\\User\\server.pem
        tags: wsl
ubuntu_pro:
    token: agent_token"""
        )

        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write("")

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        # Assert Landscape and Agent combine, with Agent taking precedence
        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )

        assert (
            "agent_test" in userdata and "agent_token" in userdata
        ), "Agent data should be present"

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_landscape_shell_script(self, m_get_linux_dist, tmpdir, paths):
        """Asserts that Pro for WSL and Landscape goes multipart"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        agent_file.write(
            """#cloud-config
landscape:
    host:
        url: hosted.com:6554
    client:
        account_name: agent_test
        url: https://hosted.com/message-system
        ping_url: https://hosted.com/ping
        ssl_public_key: C:\\Users\\User\\server.pem
        tags: wsl
ubuntu_pro:
    token: agent_token"""
        )

        COMMAND = "echo Hello cloud-init on WSL!"
        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(f"#!/bin/sh\n{COMMAND}\n")

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        # Assert Landscape and Agent combine, with Agent taking precedence
        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )

        assert (
            "agent_test" in userdata and "agent_token" in userdata
        ), "Agent data should be present"

        shell_script = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/x-shellscript"
        )

        assert COMMAND in shell_script

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_with_landscape_no_tags(self, m_get_linux_dist, tmpdir, paths):
        """Validates the Pro For WSL default Landscape tags are applied"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        agent_file.write(
            """#cloud-config
landscape:
    host:
        url: landscape.canonical.com:6554
    client:
        account_name: agenttest
        url: https://landscape.canonical.com/message-system
        ping_url: https://landscape.canonical.com/ping
        tags: wsl
ubuntu_pro:
    token: testtoken"""
        )
        # Set up some Landscape provided user data without tags
        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(
            """#cloud-config
landscape:
  client:
    account_name: landscapetest
package_update: true"""
        )

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )

        assert (
            "tags: wsl" in userdata
        ), "Landscape computer tags should match UP4W agent's data defaults"

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_with_no_tags_at_all(self, m_get_linux_dist, tmpdir, paths):
        """Asserts the DS still works if there are no Landscape tags at all"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO

        user_file = tmpdir.join(".cloud-init", "ubuntu-24.04.user-data")
        user_file.dirpath().mkdir()
        user_file.write("#cloud-config\nwrite_files:\n- path: /etc/wsl.conf")

        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        # Make sure we don't crash if there are no tags anywhere.
        agent_file.write(
            """#cloud-config
ubuntu_pro:
    token: up4w_token"""
        )
        # Set up some Landscape provided user data without tags
        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(
            """#cloud-config
landscape:
  client:
    account_name: landscapetest
package_update: true"""
        )

        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )
        assert "landscapetest" in userdata
        assert "up4w_token" in userdata
        assert "tags" not in userdata

    @mock.patch("cloudinit.util.get_linux_distro")
    def test_with_no_client_subkey(self, m_get_linux_dist, tmpdir, paths):
        """Validates the DS works without the landscape.client subkey"""

        m_get_linux_dist.return_value = SAMPLE_LINUX_DISTRO
        ubuntu_pro_tmp = tmpdir.join(".ubuntupro", ".cloud-init")
        os.makedirs(ubuntu_pro_tmp, exist_ok=True)

        agent_file = ubuntu_pro_tmp.join("agent.yaml")
        # Make sure we don't crash if there is no client subkey.
        # (That would be a bug in the agent as there is no other config
        # value for landscape outside of landscape.client, so I'm making up
        # some non-sense keys just to make sure we won't crash)
        agent_file.write(
            """#cloud-config
landscape:
    server:
        port: 6554
ubuntu_pro:
    token: up4w_token"""
        )

        landscape_file = ubuntu_pro_tmp.join("%s.user-data" % INSTANCE_NAME)
        landscape_file.write(
            """#cloud-config
landscape:
  client:
    account_name: landscapetest
package_update: true"""
        )
        # Run the datasource
        ds = wsl.DataSourceWSL(
            sys_cfg=SAMPLE_CFG, distro=get_distro("ubuntu"), paths=paths
        )

        assert ds.get_data() is True
        ud = ds.get_userdata()

        assert ud is not None
        userdata = join_payloads_from_content_type(
            cast(MIMEMultipart, ud), "text/cloud-config"
        )
        assert "landscapetest" not in userdata
        assert (
            "port: 6554" in userdata
        ), "agent data should override the entire landscape config."

        assert "up4w_token" in userdata



import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.util import verify_clean_boot

USER_DATA = """\
#cloud-config
runcmd:
 - pkill cloud-init
"""


@pytest.mark.user_data(USER_DATA)
def test_no_warnings(client: IntegrationInstance):
    """Test that the signal handler does not log errors when suppressed."""
    verify_clean_boot(client)
    log = client.read_from_file("/var/log/cloud-init.log")
    assert "Received signal 15 resulting in exit" in log



# Author: Mathieu Corbin <mathieu.corbin@exoscale.com>
# Author: Christopher Glass <christopher.glass@exoscale.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
import os

import requests
import responses

from cloudinit import helpers, util
from cloudinit.sources.DataSourceExoscale import (
    API_VERSION,
    METADATA_URL,
    PASSWORD_SERVER_PORT,
    DataSourceExoscale,
    get_password,
    read_metadata,
)
from tests.unittests.helpers import mock

TEST_PASSWORD_URL = "{}:{}/{}/".format(
    METADATA_URL, PASSWORD_SERVER_PORT, API_VERSION
)

TEST_METADATA_URL = "{}/{}/meta-data/".format(METADATA_URL, API_VERSION)

TEST_USERDATA_URL = "{}/{}/user-data".format(METADATA_URL, API_VERSION)


class TestDatasourceExoscale:
    @responses.activate
    def test_password_saved(self):
        """The password is not set when it is not found
        in the metadata service."""
        responses.add(responses.GET, TEST_PASSWORD_URL, body="saved_password")
        assert not get_password()

    @responses.activate
    def test_password_empty(self):
        """No password is set if the metadata service returns
        an empty string."""
        responses.add(responses.GET, TEST_PASSWORD_URL, body="")
        assert not get_password()

    @responses.activate
    def test_password(self):
        """The password is set to what is found in the metadata
        service."""
        expected_password = "p@ssw0rd"
        responses.add(responses.GET, TEST_PASSWORD_URL, body=expected_password)
        password = get_password()
        assert expected_password == password

    def test_activate_removes_set_passwords_semaphore(self, tmp_path):
        """Allow set_passwords to run every boot by removing the semaphore."""
        path = helpers.Paths({"cloud_dir": str(tmp_path)})
        sem_dir = str(tmp_path / "instance/sem")
        util.ensure_dir(sem_dir)
        sem_file = os.path.join(sem_dir, "config_set_passwords")
        with open(sem_file, "w") as stream:
            stream.write("")
        ds = DataSourceExoscale({}, None, path)
        ds.activate(None, None)
        assert not os.path.exists(sem_file)

    @responses.activate
    def test_get_data(self, tmp_path):
        """The datasource conforms to expected behavior when supplied
        full test data."""
        path = helpers.Paths({"run_dir": str(tmp_path)})
        ds = DataSourceExoscale({}, None, path)
        ds.ds_detect = lambda: True
        expected_password = "p@ssw0rd"
        expected_id = "12345"
        expected_hostname = "myname"
        expected_userdata = "#cloud-config"
        responses.add(responses.GET, TEST_USERDATA_URL, body=expected_userdata)
        responses.add(responses.GET, TEST_PASSWORD_URL, body=expected_password)
        responses.add(
            responses.GET,
            TEST_METADATA_URL,
            body="instance-id\nlocal-hostname",
        )
        responses.add(
            responses.GET,
            "{}local-hostname".format(TEST_METADATA_URL),
            body=expected_hostname,
        )
        responses.add(
            responses.GET,
            "{}instance-id".format(TEST_METADATA_URL),
            body=expected_id,
        )
        assert ds._check_and_get_data()
        assert ds.userdata_raw.decode("utf-8") == "#cloud-config"
        assert ds.metadata == {
            "instance-id": expected_id,
            "local-hostname": expected_hostname,
        }
        assert ds.get_config_obj() == {
            "ssh_pwauth": True,
            "password": expected_password,
            "chpasswd": {
                "expire": False,
            },
        }

    @responses.activate
    def test_get_data_saved_password(self, tmp_path):
        """The datasource conforms to expected behavior when saved_password is
        returned by the password server."""
        path = helpers.Paths({"run_dir": str(tmp_path)})
        ds = DataSourceExoscale({}, None, path)
        ds.ds_detect = lambda: True
        expected_answer = "saved_password"
        expected_id = "12345"
        expected_hostname = "myname"
        expected_userdata = "#cloud-config"
        responses.add(responses.GET, TEST_USERDATA_URL, body=expected_userdata)
        responses.add(responses.GET, TEST_PASSWORD_URL, body=expected_answer)
        responses.add(
            responses.GET,
            TEST_METADATA_URL,
            body="instance-id\nlocal-hostname",
        )
        responses.add(
            responses.GET,
            "{}local-hostname".format(TEST_METADATA_URL),
            body=expected_hostname,
        )
        responses.add(
            responses.GET,
            "{}instance-id".format(TEST_METADATA_URL),
            body=expected_id,
        )
        assert ds._check_and_get_data()
        assert ds.userdata_raw.decode("utf-8") == "#cloud-config"
        assert ds.metadata == {
            "instance-id": expected_id,
            "local-hostname": expected_hostname,
        }
        assert ds.get_config_obj() == {}

    @responses.activate
    def test_get_data_no_password(self, tmp_path):
        """The datasource conforms to expected behavior when no password is
        returned by the password server."""
        path = helpers.Paths({"run_dir": str(tmp_path)})
        ds = DataSourceExoscale({}, None, path)
        ds.ds_detect = lambda: True
        expected_answer = ""
        expected_id = "12345"
        expected_hostname = "myname"
        expected_userdata = "#cloud-config"
        responses.add(responses.GET, TEST_USERDATA_URL, body=expected_userdata)
        responses.add(responses.GET, TEST_PASSWORD_URL, body=expected_answer)
        responses.add(
            responses.GET,
            TEST_METADATA_URL,
            body="instance-id\nlocal-hostname",
        )
        responses.add(
            responses.GET,
            "{}local-hostname".format(TEST_METADATA_URL),
            body=expected_hostname,
        )
        responses.add(
            responses.GET,
            "{}instance-id".format(TEST_METADATA_URL),
            body=expected_id,
        )
        assert ds._check_and_get_data()
        assert ds.userdata_raw.decode("utf-8") == "#cloud-config"
        assert ds.metadata == {
            "instance-id": expected_id,
            "local-hostname": expected_hostname,
        }
        assert ds.get_config_obj() == {}

    @responses.activate
    @mock.patch("cloudinit.sources.DataSourceExoscale.get_password")
    def test_read_metadata_when_password_server_unreachable(self, m_password):
        """The read_metadata function returns partial results in case the
        password server (only) is unreachable."""
        expected_id = "12345"
        expected_hostname = "myname"
        expected_userdata = "#cloud-config"

        m_password.side_effect = requests.Timeout("Fake Connection Timeout")
        responses.add(responses.GET, TEST_USERDATA_URL, body=expected_userdata)
        responses.add(
            responses.GET,
            TEST_METADATA_URL,
            body="instance-id\nlocal-hostname",
        )
        responses.add(
            responses.GET,
            "{}local-hostname".format(TEST_METADATA_URL),
            body=expected_hostname,
        )
        responses.add(
            responses.GET,
            "{}instance-id".format(TEST_METADATA_URL),
            body=expected_id,
        )

        result = read_metadata()

        assert result.get("password") is None
        assert result.get("user-data").decode("utf-8") == expected_userdata

    def test_non_viable_platform(self, tmp_path):
        """The datasource fails fast when the platform is not viable."""
        path = helpers.Paths({"run_dir": str(tmp_path)})
        ds = DataSourceExoscale({}, None, path)
        ds.ds_detect = lambda: False
        assert not ds._check_and_get_data()



# This file is part of cloud-init. See LICENSE file for license information.
from cloudinit.distros.parsers import hostname

BASE_HOSTNAME = """
# My super-duper-hostname

blahblah

"""
BASE_HOSTNAME = BASE_HOSTNAME.strip()


class TestHostnameHelper:
    def test_parse_same(self):
        hn = hostname.HostnameConf(BASE_HOSTNAME)
        assert str(hn).strip() == BASE_HOSTNAME
        assert hn.hostname == "blahblah"

    def test_no_adjust_hostname(self):
        hn = hostname.HostnameConf(BASE_HOSTNAME)
        prev_name = hn.hostname
        hn.set_hostname("")
        assert hn.hostname == prev_name

    def test_adjust_hostname(self):
        hn = hostname.HostnameConf(BASE_HOSTNAME)
        prev_name = hn.hostname
        assert prev_name == "blahblah"
        hn.set_hostname("bbbbd")
        assert hn.hostname == "bbbbd"
        expected_out = """
# My super-duper-hostname

bbbbd
"""
        assert str(hn).strip() == expected_out.strip()



# This file is part of cloud-init. See LICENSE file for license information.

import logging
from typing import MutableMapping

from cloudinit import net, url_helper, util
from cloudinit.sources.helpers import ec2

LOG = logging.getLogger(__name__)


def get_instance_meta_data(
    api_version="latest",
    metadata_address="http://100.100.100.200",
    ssl_details=None,
    timeout=5,
    retries=5,
    headers_cb=None,
    headers_redact=None,
    exception_cb=None,
):
    ud_url = url_helper.combine_url(metadata_address, api_version)
    ud_url = url_helper.combine_url(ud_url, "meta-data/all")
    response = url_helper.read_file_or_url(
        ud_url,
        ssl_details=ssl_details,
        timeout=timeout,
        retries=retries,
        exception_cb=exception_cb,
        headers_cb=headers_cb,
        headers_redact=headers_redact,
    )
    meta_data_raw: object = util.load_json(response.contents)

    # meta_data_raw is a json object with the following format get
    # by`meta-data/all`
    # {
    #   "sub-private-ipv4-list": "",
    #   "dns-conf": {
    #     "nameservers": "100.100.2.136\r\n100.100.2.138"
    #   },
    #   "zone-id": "cn-hangzhou-i",
    #   "instance": {
    #     "instance-name": "aliyun_vm_test",
    #     "instance-type": "ecs.g7.xlarge"
    #   },
    #   "disks": {
    #     "bp1cikh4di1xxxx": {
    #       "name": "disk_test",
    #       "id": "d-bp1cikh4di1lf7pxxxx"
    #     }
    #   },
    #   "instance-id": "i-bp123",
    #   "eipv4": "47.99.152.7",
    #   "private-ipv4": "192.168.0.9",
    #   "hibernation": {
    #     "configured": "false"
    #   },
    #   "vpc-id": "vpc-bp1yeqg123",
    #   "mac": "00:16:3e:30:3e:ca",
    #   "source-address": "http://mirrors.cloud.aliyuncs.com",
    #   "vswitch-cidr-block": "192.168.0.0/24",
    #   "network": {
    #     "interfaces": {
    #       "macs": {
    #         "00:16:3e:30:3e:ca": {
    #           "vpc-cidr-block": "192.168.0.0/16",
    #           "netmask": "255.255.255.0"
    #         }
    #       }
    #     }
    #   },
    #   "network-type": "vpc",
    #   "hostname": "aliyun_vm_test",
    #   "region-id": "cn-hangzhou",
    #   "ntp-conf": {
    #     "ntp-servers": "ntp1.aliyun.com\r\nntp2.aliyun.com"
    #   },
    # }
    # Note: For example, in the values of dns conf: the `nameservers`
    # key is a string, the format is the same as the response from the
    # `meta-data/dns-conf/nameservers` endpoint. we use the same
    # serialization method to ensure consistency between
    # the two methods (directory tree and json path).
    def _process_dict_values(d):
        if isinstance(d, dict):
            return {k: _process_dict_values(v) for k, v in d.items()}
        elif isinstance(d, list):
            return [_process_dict_values(item) for item in d]
        else:
            return ec2.MetadataLeafDecoder()("", d)

    return _process_dict_values(meta_data_raw)


def get_instance_data(
    api_version="latest",
    metadata_address="http://100.100.100.200",
    ssl_details=None,
    timeout=5,
    retries=5,
    headers_cb=None,
    headers_redact=None,
    exception_cb=None,
    item_name=None,
):
    ud_url = url_helper.combine_url(metadata_address, api_version)
    ud_url = url_helper.combine_url(ud_url, item_name)
    data = b""
    support_items_list = ["user-data", "vendor-data"]
    if item_name not in support_items_list:
        LOG.error(
            "aliyun datasource not support the item  %s",
            item_name,
        )
        return data
    try:
        response = url_helper.read_file_or_url(
            ud_url,
            ssl_details=ssl_details,
            timeout=timeout,
            retries=retries,
            exception_cb=exception_cb,
            headers_cb=headers_cb,
            headers_redact=headers_redact,
        )
        data = response.contents
    except Exception:
        util.logexc(LOG, "Failed fetching %s from url %s", item_name, ud_url)
    return data


def convert_ecs_metadata_network_config(
    network_md,
    macs_to_nics=None,
    fallback_nic=None,
    full_network_config=True,
):
    """Convert ecs metadata to network config version 2 data dict.

    @param: network_md: 'network' portion of ECS metadata.
    generally formed as {"interfaces": {"macs": {}} where
    'macs' is a dictionary with mac address as key:
    @param: macs_to_nics: Optional dict of mac addresses and nic names. If
    not provided, get_interfaces_by_mac is called to get it from the OS.
    @param: fallback_nic: Optionally provide the primary nic interface name.
    This nic will be guaranteed to minimally have a dhcp4 configuration.
    @param: full_network_config: Boolean set True to configure all networking
    presented by IMDS. This includes rendering secondary IPv4 and IPv6
    addresses on all NICs and rendering network config on secondary NICs.
    If False, only the primary nic will be configured and only with dhcp
    (IPv4/IPv6).

    @return A dict of network config version 2 based on the metadata and macs.
    """
    netcfg: MutableMapping = {"version": 2, "ethernets": {}}
    if not macs_to_nics:
        macs_to_nics = net.get_interfaces_by_mac()
    macs_metadata = network_md["interfaces"]["macs"]

    if not full_network_config:
        for mac, nic_name in macs_to_nics.items():
            if nic_name == fallback_nic:
                break
        dev_config: MutableMapping = {
            "dhcp4": True,
            "dhcp6": False,
            "match": {"macaddress": mac.lower()},
            "set-name": nic_name,
        }
        nic_metadata = macs_metadata.get(mac)
        if nic_metadata.get("ipv6s"):  # Any IPv6 addresses configured
            dev_config["dhcp6"] = True
        netcfg["ethernets"][nic_name] = dev_config
        return netcfg
    nic_name_2_mac_map = dict()
    for mac, nic_name in macs_to_nics.items():
        nic_metadata = macs_metadata.get(mac)
        if not nic_metadata:
            continue  # Not a physical nic represented in metadata
        nic_name_2_mac_map[nic_name] = mac

    # sorted by nic_name
    ordered_nic_name_list = sorted(
        nic_name_2_mac_map.keys(), key=net.natural_sort_key
    )
    for nic_idx, nic_name in enumerate(ordered_nic_name_list):
        nic_mac = nic_name_2_mac_map[nic_name]
        nic_metadata = macs_metadata.get(nic_mac)
        dhcp_override = {"route-metric": (nic_idx + 1) * 100}
        dev_config = {
            "dhcp4": True,
            "dhcp4-overrides": dhcp_override,
            "dhcp6": False,
            "match": {"macaddress": nic_mac.lower()},
            "set-name": nic_name,
        }
        if nic_metadata.get("ipv6s"):  # Any IPv6 addresses configured
            dev_config["dhcp6"] = True
            dev_config["dhcp6-overrides"] = dhcp_override

        netcfg["ethernets"][nic_name] = dev_config
    # Remove route-metric dhcp overrides and routes / routing-policy if only
    # one nic configured
    if len(netcfg["ethernets"]) == 1:
        for nic_name in netcfg["ethernets"].keys():
            netcfg["ethernets"][nic_name].pop("dhcp4-overrides")
            netcfg["ethernets"][nic_name].pop("dhcp6-overrides", None)
            netcfg["ethernets"][nic_name].pop("routes", None)
            netcfg["ethernets"][nic_name].pop("routing-policy", None)
    return netcfg



import random
import signal
import socket
import time
from threading import Thread
from unittest import mock

from cloudinit import socket as ci_socket


class Sync:
    """A device to send and receive synchronization messages

    Creating an instance of the device sends a b"start"
    """

    def __init__(self, name: str, path: str):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(f"{path}/share/{name}.sock")
        self.sock.sendall(b"start")

    def receive(self):
        """receive 5 bytes from the socket"""
        received = self.sock.recv(4096)
        self.sock.close()
        return received


class Timeout:
    """A utility which may be used to verify that a timeout occurs

    TimeoutError is raised on successful timeout.

    Create a signal handler and use signal.alarm to verify that the
    timeout occurred.
    """

    def handle_timeout(self, *_):
        raise TimeoutError()

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        # 1 second is, unfortunately, the minimum
        signal.alarm(1)

    def __exit__(self, *_):
        signal.alarm(0)


def test_all_stages_times_out(tmp_path, socket_attrs):
    """Verify that no "start" makes the protocol block"""
    with mock.patch.object(
        ci_socket, "DEFAULT_RUN_DIR", tmp_path
    ), mock.patch.object(ci_socket, "sd_notify"), mock.patch.object(
        ci_socket.os, "isatty", return_value=False
    ), mock.patch.object(
        ci_socket.sys.stdin, "fileno"
    ):
        sync = ci_socket.SocketSync("first")

        try:
            with Timeout():
                # this should block for 1 second
                with sync("first"):
                    pass
        except TimeoutError:
            # success is a timeout
            pass
        else:
            raise AssertionError("Expected the thing to timeout!")


def test_all_stages(tmp_path, socket_attrs):
    """Verify that a socket can store "start" messages

    After a socket has been been bound but before it has started listening
    """
    expected = "echo 'Completed socket interaction for boot stage {}'; exit 0;"
    with mock.patch.object(
        ci_socket, "DEFAULT_RUN_DIR", tmp_path
    ), mock.patch.object(ci_socket, "sd_notify"), mock.patch.object(
        ci_socket.os, "isatty", return_value=False
    ), mock.patch.object(
        ci_socket.sys.stdin, "fileno"
    ):
        sync = ci_socket.SocketSync("first", "second", "third")

        # send all three syncs to the sockets
        first = Sync("first", tmp_path)
        second = Sync("second", tmp_path)
        third = Sync("third", tmp_path)

        # "wait" on the first sync event
        with sync("first"):
            pass

        # check that the first sync returned
        assert expected.format("first").encode() == first.receive()
        # "wait" on the second sync event
        with sync("second"):
            pass
        # check that the second sync returned
        assert expected.format("second").encode() == second.receive()
        # "wait" on the third sync event
        with sync("third"):
            pass
        # check that the third sync returned
        assert expected.format("third").encode() == third.receive()


def test_all_stages_threaded(tmp_path, socket_attrs):
    """Verify that arbitrary "start" order works"""

    # in milliseconds
    max_sleep = 100
    # initialize random number generator
    random.seed(time.time())
    expected = "echo 'Completed socket interaction for boot stage {}'; exit 0;"
    sync_storage = {}

    def syncer(index: int, name: str):
        """sleep for 0-100ms then send a sync notification

        this allows sync order to be arbitrary
        """
        time.sleep(0.001 * random.randint(0, max_sleep))
        sync_storage[index] = Sync(name, tmp_path)

    with mock.patch.object(
        ci_socket, "DEFAULT_RUN_DIR", tmp_path
    ), mock.patch.object(ci_socket, "sd_notify"), mock.patch.object(
        ci_socket.os, "isatty", return_value=False
    ), mock.patch.object(
        ci_socket.sys.stdin, "fileno"
    ):

        sync = ci_socket.SocketSync(
            "first", "second", "third", "fourth", "fifth"
        )

        for i, name in {
            1: "first",
            2: "second",
            3: "third",
            4: "fourth",
            5: "fifth",
        }.items():
            t = Thread(target=syncer, args=(i, name))
            t.run()

        # wait on the first sync event
        with sync("first"):
            pass

        # check that the first sync returned
        assert expected.format("first").encode() == sync_storage[1].receive()

        # wait on the second sync event
        with sync("second"):
            pass

        # check that the second sync returned
        assert expected.format("second").encode() == sync_storage[2].receive()

        # wait on the third sync event
        with sync("third"):
            pass

        # check that the third sync returned
        assert expected.format("third").encode() == sync_storage[3].receive()
        with sync("fourth"):
            pass

        # check that the fourth sync returned
        assert expected.format("fourth").encode() == sync_storage[4].receive()

        with sync("fifth"):
            pass

        # check that the fifth sync returned
        assert expected.format("fifth").encode() == sync_storage[5].receive()


def test_all_stages_exception(tmp_path, socket_attrs):
    """Verify that exceptions log messages produce a valid warning message"""
    with mock.patch.object(
        ci_socket, "DEFAULT_RUN_DIR", tmp_path
    ), mock.patch.object(ci_socket, "sd_notify"), mock.patch.object(
        ci_socket.os, "isatty", return_value=False
    ), mock.patch.object(
        ci_socket.sys.stdin, "fileno"
    ):
        sync = ci_socket.SocketSync("first", "second", "third")

        # send all three syncs to the sockets
        first = Sync("first", tmp_path)

        # "wait" on the first sync event
        with sync("first"):
            # verify that an exception in context doesn't raise
            1 / 0  # pylint: disable=W0104

        assert (
            b"echo 'fatal error, run \"systemctl status cloud-init-main."
            b'service" and "cloud-init status --long" for '
            b"more details'; exit 1;" == first.receive()
        )
